import typing

import numpy as np
import pandas as pd

from ._alifestd_categorize_triplet_asexual import (
    alifestd_categorize_triplet_asexual,
)
from ._alifestd_find_leaf_ids import alifestd_find_leaf_ids
from ._alifestd_find_mrca_id_asexual import alifestd_find_mrca_id_asexual


def alifestd_sample_triplet_comparisons_asexual(
    first_df: pd.DataFrame,
    second_df: pd.DataFrame,
    taxon_label_key: str,
    n: int = 1000,
    progress_wrap: typing.Callable = lambda x: x,
    mutate: bool = False,
) -> pd.DataFrame:
    """Sample triplet comparisons between two asexual phylogenetic trees in
    alife standard form, creating a DataFrame with the triplet categorizations
    and comparison results as well as corresponding data from MRCA row within
    the first tree.

    The MRCA row corresponds to the most recent common ancestor of two of the
    three taxa in the triplet.

    Parameters
    ----------
    first_df : pd.DataFrame
        The DataFrame representing the first phylogenetic tree.
    second_df : pd.DataFrame
        The DataFrame representing the second phylogenetic tree.
    taxon_label_key : str
        The key in the DataFrame to identify the taxon labels.
    n : int, default 1000
        The number of samples to take.

        Corresponds to number of rows in the returned DataFrame.
    progress_wrap : typing.Callable, optional
        Pass tqdm or equivalent to display a progress bar.
    mutate : bool, default False
        If True, allows mutation of input DataFrames.

    Returns
    -------
    pd.DataFrame
        A DataFrame with rows corresponding to sampled triplet comparisons and
        the following columns:
        - "triplet code, {first,second}": the categorization of the triplet in
          the first or second tree.
        - "triplet match, {lax,lax/strict,strict,strict/lax}": whether the
          triplet categorizations match with differing treatment of polytomies.
        - all columns from the first tree.

    Notes
    -----
    The core comparison is done by sampling triplets of taxa, categorizing
    them, and comparing these categorizations across the two trees, taking into
    account the `strict` and `lax` parameters for handling polytomies. See
    `alifestd_categorize_triplet_asexual` for details.

    See Also
    --------
    alifestd_categorize_triplet_asexual
    alifestd_estimate_triplet_distance_asexual
    """
    if not mutate:
        first_df = first_df.copy()
        second_df = second_df.copy()

    first_df = first_df.reset_index(drop=True).set_index("id", drop=False)
    second_df = second_df.reset_index(drop=True).set_index("id", drop=False)

    first_leaf_ids = alifestd_find_leaf_ids(first_df)
    first_taxon_labels = first_df.loc[first_leaf_ids, taxon_label_key]
    second_leaf_ids = alifestd_find_leaf_ids(second_df)
    second_taxon_labels = second_df.loc[second_leaf_ids, taxon_label_key]

    assert first_taxon_labels.is_unique
    assert second_taxon_labels.is_unique
    assert (
        first_taxon_labels.sort_values().to_numpy()
        == second_taxon_labels.sort_values().to_numpy()
    ).all()

    if len(first_leaf_ids) < 3:
        raise ValueError("Too few leaves to sample any triplets.")

    def sample_triplet_comparison() -> dict:
        """Sample details of one triplet comparison between trees."""
        labels = np.random.choice(first_taxon_labels, 3, replace=False)
        first_triplet_ids = [
            first_df.index[first_df[taxon_label_key] == label].item()
            for label in labels
        ]
        second_triplet_ids = [
            second_df.index[second_df[taxon_label_key] == label].item()
            for label in labels
        ]
        assert 3 == len(first_triplet_ids) == len(second_triplet_ids)

        cat1 = alifestd_categorize_triplet_asexual(
            first_df, first_triplet_ids, mutate=True
        )
        cat2 = alifestd_categorize_triplet_asexual(
            second_df, second_triplet_ids, mutate=True
        )

        first_sampled_leaf_pair = np.random.choice(
            first_triplet_ids, 2, replace=False
        )
        sampled_mrca_id = alifestd_find_mrca_id_asexual(
            first_df, first_sampled_leaf_pair, mutate=True
        )
        polytomy = -1
        return {
            "triplet code, first": cat1,
            "triplet code, second": cat2,
            "triplet match, lax": (
                (cat1 == cat2) or (cat1 == polytomy) or (cat2 == polytomy)
            ),
            "triplet match, lax/strict": (cat1 == cat2) or (cat1 == polytomy),
            "triplet match, strict": cat1 == cat2,
            "triplet match, strict/lax": (cat1 == cat2) or (cat2 == polytomy),
            **first_df.loc[sampled_mrca_id].to_dict(),
        }

    return pd.DataFrame.from_records(
        [sample_triplet_comparison() for _ in progress_wrap(range(n))],
    )



import sys
import typing

import interval_search as inch
import numpy as np
import opytional as opyt

from ..stratum_retention_algorithms._detail import PolicySpecBase
from ._detail import policy_evaluator_t


class PropertyAtLeastParameterizer:
    """Parameterizes so evaluated property meets or exceeds a target value."""

    _target_value: typing.Union[float, int]
    _policy_evaluator: policy_evaluator_t
    _param_lower_bound: int
    _param_upper_bound: typing.Optional[int]

    def __init__(
        self: "PropertyAtLeastParameterizer",
        target_value: typing.Union[float, int],
        policy_evaluator: policy_evaluator_t,
        param_lower_bound: int = 0,
        param_upper_bound: typing.Optional[int] = sys.maxsize,
    ) -> None:
        """Init functor with parameterization requirements.

        Parameters
        ----------
        target_value : float or int
            The threshold property value to parameterize the policy to meet.
        policy_evaluator : int
            Functor that evaluates a policy at a particular parameter value.
        param_lower_bound : int
            Lower value on the range of parameter values to search, inclusive.
        param_upper_bound : int, optional
            Upper bound on the range of parameter values to search, inclusive.
            If None, no upper bound is used.
        """
        self._target_value = target_value
        self._policy_evaluator = policy_evaluator
        self._param_lower_bound = param_lower_bound
        self._param_upper_bound = param_upper_bound

    def __call__(
        self: "PropertyAtLeastParameterizer",
        policy_t: typing.Type,
    ) -> typing.Optional[PolicySpecBase]:
        """Solve for policy spec satisfying parameterization requirements."""
        policy_factory = self._policy_evaluator._policy_param_focalizer(
            policy_t,
        )
        try:
            res = self._try_calc_parameter(policy_t)
            if res is not None:
                assert self._param_lower_bound <= res
                if self._param_upper_bound is not None:
                    assert res <= self._param_upper_bound
            return opyt.apply_if(
                res,
                lambda x: policy_factory(x).GetSpec(),
            )
        except (MemoryError, OverflowError, RecursionError):
            return None

    def _try_calc_parameter(
        self: "PropertyAtLeastParameterizer",
        policy_t: typing.Type,
    ) -> typing.Optional[int]:
        lb = self._param_lower_bound
        ub = self._param_upper_bound
        thresh = self._target_value

        def eval_at_param(val: int) -> typing.Union[float, int]:
            return self._policy_evaluator(policy_t, val)

        next_diff_param = inch.interval_search(
            lambda p: eval_at_param(p) != eval_at_param(lb),
            lower_bound=lb,
            upper_bound=ub,
        )
        sign = opyt.apply_if(
            next_diff_param,
            lambda x: np.sign(eval_at_param(x) - eval_at_param(lb)),
        )
        assert sign != 0

        if sign is None or sign == 1:
            # if all parameters in search range evaluate to same value
            # (sign is None) or values increase with parameter (sign == 1)
            # forward search for parameter, if any, that satisfies
            # parameterization requirement
            return inch.interval_search(
                lambda p: eval_at_param(p) >= thresh,
                lower_bound=lb,
                upper_bound=ub,
            )
        elif eval_at_param(lb) >= thresh:
            # if value decreases with parameter and the lowest value satisfies
            # parameterization requirement, forward search for last parameter
            # that satisfies parameterization requirement
            assert sign == -1
            try:
                res = inch.interval_search(
                    # successor fails requirement
                    lambda p: eval_at_param(p + 1) < thresh,
                    lower_bound=lb,
                    # override upper bound to avoid unnecessary infinite search
                    # b/c doubling_search is implemented non-recursively
                    upper_bound=opyt.or_value(ub, sys.maxsize) - 1,
                )
                if res is not None:
                    return res
            except (MemoryError, OverflowError, RecursionError):
                pass
            # looking for parameter with successor that fails requirement
            # failed
            # try again, looking for any parameter that minimally satisfies
            # requirement
            try:
                res = inch.interval_search(
                    lambda p: eval_at_param(p) == thresh,
                    lower_bound=lb,
                    # override upper bound to avoid unnecessary infinite search
                    # b/c doubling_search is implemented non-recursively
                    upper_bound=opyt.or_value(ub, sys.maxsize),
                )
                if res is not None:
                    return res
            except (MemoryError, OverflowError, RecursionError):
                pass
            # all searchable parameters sastisfy requirement, so
            # arbitrarily pick
            return opyt.or_value(ub, lb)
        else:
            assert sign == -1
            assert eval_at_param(lb) < thresh
            return None

    def __repr__(self: "PropertyAtLeastParameterizer") -> str:
        return f"""{
            PropertyAtLeastParameterizer.__qualname__
        }(target_value={
            self._target_value
        !r}, policy_evaluator={
            self._policy_evaluator
        !r}, param_lower_bound={
            self._param_lower_bound
        !r}, param_upper_bound={
            self._param_upper_bound
        !r})"""

    def __str__(self: "PropertyAtLeastParameterizer") -> str:
        title = "At Least Parameterizer"
        return f"""{
            title
        } (target value: {
            self._target_value
        }, evaluator: {
            self._policy_evaluator
        }, param lower bound: {
            self._param_lower_bound
        }, param upper bound: {
            opyt.or_value(self._param_upper_bound, 'inf')
        })"""



import typing

from ...._detail import PolicyCouplerBase
from ..._PolicySpec import PolicySpec
from ..._impl import calc_provided_uncertainty


class GenDropRanks:
    """Functor to implement the depth-proportional resolution stratum retention
    policy, for use with HereditaryStratigraphicColumn.

    This functor enacts the depth-proportional resolution policy by specifying
    the set of strata ranks that should be purged from a hereditary
    stratigraphic column when the nth stratum is deposited.
    """

    def __init__(
        self: "GenDropRanks",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "GenDropRanks", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "GenDropRanks",
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        retained_ranks: typing.Optional[typing.Iterable[int]],
    ) -> typing.Iterator[int]:
        """Decide which strata within the stratagraphic column should be purged.

        Every time a new stratum is deposited, this method is called to
        determine which strata should be purged. All strata at ranks yielded
        from this functor are immediately purged from the column, meaning that
        for a stratum to persist it must not be yielded by this functor each
        and every time a new stratum is deposited.

        Parameters
        ----------
        policy: Policy
            Policy this functor enacts.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.
        retained_ranks : iterator over int
            An iterator over ranks of strata currently retained within the
            hereditary stratigraphic column. Not used in this functor.

        Returns
        -------
        iterator over int
                The ranks of strata that should be purged from the hereditary
            stratigraphic column at this deposition step.

        See Also
        --------
        depth_proportional_resolution_algo:
            For details on the rationale, implementation, and guarantees of the
            depth-proportional resolution stratum retention policy.
        """
        spec = policy.GetSpec()
        resolution = spec.GetDepthProportionalResolution()

        # until sufficient strata have been deposited to reach target resolution
        # don't remove any strata
        if num_stratum_depositions_completed <= resolution:
            return

        # newest stratum is in-progress deposition
        # that will occupy rank num_stratum_depositions_completed
        second_newest_stratum_rank = num_stratum_depositions_completed - 1
        # +1's because of in-progress deposition
        # _calc_provided_uncertainty is from super class
        cur_provided_uncertainty = calc_provided_uncertainty(
            resolution,
            num_stratum_depositions_completed + 1,
        )
        prev_provided_uncertainty = calc_provided_uncertainty(
            resolution,
            num_stratum_depositions_completed + 1 - 1,
        )
        if cur_provided_uncertainty != prev_provided_uncertainty:
            # we just passed the threshold where the spacing between retained
            # strata could be doubled without violating our resolution guarantee
            # clean up no-longer-needed strata that bisect
            # cur_provided_uncertainty intervals
            assert prev_provided_uncertainty * 2 == cur_provided_uncertainty
            yield from range(
                prev_provided_uncertainty,
                second_newest_stratum_rank,
                cur_provided_uncertainty,
            )

        if second_newest_stratum_rank % cur_provided_uncertainty:
            # we always keep the newest stratum
            # but unless the now-second-newest stratum is needed as a waypoint
            # of the cur_provided_uncertainty intervals, get rid of it
            yield second_newest_stratum_rank



import itertools as it

import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.95, 0.88],
)
def test_does_have_any_common_ancestor_specimen(
    retention_policy, differentia_width, confidence_level
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.does_have_any_common_ancestor(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column2),
        confidence_level=confidence_level,
    ) == hstrat.does_have_any_common_ancestor(
        column, column2, confidence_level=confidence_level
    )

    assert hstrat.does_have_any_common_ancestor(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column),
        confidence_level=confidence_level,
    ) == hstrat.does_have_any_common_ancestor(
        column, column, confidence_level=confidence_level
    )

    assert hstrat.does_have_any_common_ancestor(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(child1),
        confidence_level=confidence_level,
    ) == hstrat.does_have_any_common_ancestor(
        column, child1, confidence_level=confidence_level
    )

    assert hstrat.does_have_any_common_ancestor(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        confidence_level=confidence_level,
    ) == hstrat.does_have_any_common_ancestor(
        child1, child2, confidence_level=confidence_level
    )

    child1.DepositStrata(10)
    assert hstrat.does_have_any_common_ancestor(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        confidence_level=confidence_level,
    ) == hstrat.does_have_any_common_ancestor(
        child1, child2, confidence_level=confidence_level
    )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_HasAnyCommonAncestorWith(retention_policy, ordered_store):
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    first_copy = first.Clone()
    second = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    def assert_validity():
        assert hstrat.does_have_any_common_ancestor(first, first)
        assert hstrat.does_have_any_common_ancestor(first_copy, first_copy)
        assert hstrat.does_have_any_common_ancestor(first, first_copy)
        assert hstrat.does_have_any_common_ancestor(first_copy, first)

        assert not hstrat.does_have_any_common_ancestor(second, first)
        assert not hstrat.does_have_any_common_ancestor(first, second)
        assert not hstrat.does_have_any_common_ancestor(second, first_copy)
        assert not hstrat.does_have_any_common_ancestor(first_copy, second)

    assert_validity()
    first.DepositStratum()
    assert_validity()
    second.DepositStratum()
    assert_validity()
    first.DepositStratum()
    assert_validity()

    for __ in range(100):
        first_copy.DepositStratum()
        assert_validity()


def test_HasAnyCommonAncestorWith_narrow():
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    assert hstrat.does_have_any_common_ancestor(c1, c2) is None
    assert hstrat.does_have_any_common_ancestor(c2, c1) is None

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    assert hstrat.does_have_any_common_ancestor(c1, c2) is False
    assert hstrat.does_have_any_common_ancestor(c2, c1) is False

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    c2 = c1.Clone()
    assert hstrat.does_have_any_common_ancestor(c1, c2) is None
    assert hstrat.does_have_any_common_ancestor(c2, c1) is None

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    c2 = c1.Clone()
    assert hstrat.does_have_any_common_ancestor(c1, c2) is True
    assert hstrat.does_have_any_common_ancestor(c2, c1) is True

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    for __ in range(100):
        c1.DepositStratum()
    c2 = c1.CloneDescendant()
    assert hstrat.does_have_any_common_ancestor(c1, c2) is True
    assert hstrat.does_have_any_common_ancestor(c2, c1) is True

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    for __ in range(100):
        c1.DepositStratum()
    c2 = c1.CloneDescendant()
    assert hstrat.does_have_any_common_ancestor(c1, c2) is True
    assert hstrat.does_have_any_common_ancestor(c2, c1) is True


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
def test_artifact_types_equiv(differentia_width, policy):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.does_have_any_common_ancestor(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
        ) == hstrat.does_have_any_common_ancestor(a, b)



from ......_auxiliary_lib import count_trailing_zeros


def num_to_condemn_production(
    recency_proportional_resolution: int,
    num_stratum_depositions_completed: int,
) -> int:
    """How many strata should be eliminated after
    num_stratum_depositions_completed have been deposited and the
    num_stratum_depositions_completed + 1'th deposition is in progress?

    Inscrutable bit magic translation of num_to_condemn_reference.
    """
    # protect against stray numpy integer types
    recency_proportional_resolution = int(recency_proportional_resolution)
    num_stratum_depositions_completed = int(num_stratum_depositions_completed)

    comparison_value = 2 * (recency_proportional_resolution + 1)
    binary_magnitude_diff = max(
        num_stratum_depositions_completed.bit_length()
        - comparison_value.bit_length(),
        0,
    )
    comparison_value <<= binary_magnitude_diff

    # did we shift enough to equalize magnitude of comparison value?
    assert (
        num_stratum_depositions_completed.bit_length()
        <= comparison_value.bit_length()
    )

    # add one if need to perform one final shift to be bigger
    binary_magnitude_diff += (
        num_stratum_depositions_completed >= comparison_value
    )

    return min(
        count_trailing_zeros(num_stratum_depositions_completed),
        binary_magnitude_diff,
    )



import typing

import typing_extensions

from .._auxiliary_lib import get_hstrat_version
from ..genome_instrumentation import HereditaryStratigraphicColumn
from ._col_from_records import col_from_records
from ._impl import (
    DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH,
    stringify_packed_differentia_bytes,
)
from ._policy_to_records import policy_to_records


def col_from_packet(
    packet: typing_extensions.Buffer,
    differentia_bit_width: int,
    stratum_retention_policy: typing.Callable,
    differentiae_byte_bit_order: typing.Literal["big", "little"] = "big",
    num_strata_deposited_byte_order: typing.Literal["big", "little"] = "big",
    num_strata_deposited_byte_width: int = (
        DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH
    ),
) -> HereditaryStratigraphicColumn:
    """Deserialize a `HereditaryStratigraphicColumn` from a differentia packet
    and column configuration specification information.

    Use when buffer size equals packet size.

    See Also
    --------
    col_from_packet_buffer: use when buffer size exceeds packet size.
    """

    policy_records = policy_to_records(stratum_retention_policy)
    num_strata_deposited = int.from_bytes(
        packet[:num_strata_deposited_byte_width],
        byteorder=num_strata_deposited_byte_order,
        signed=False,
    )
    differentiae_records = {
        "num_strata_deposited": num_strata_deposited,
        "differentiae": stringify_packed_differentia_bytes(
            packet[num_strata_deposited_byte_width:],
        ),
        "omits_num_padding_bits_header": True,
        "differentia_bit_width": differentia_bit_width,
        "hstrat_version": get_hstrat_version(),
    }

    return col_from_records(
        {**policy_records, **differentiae_records},
        differentiae_byte_bit_order=differentiae_byte_bit_order,
    )



import numpy as np

from ._detail import PriorBase


class GeometricPrior(PriorBase):
    """Enacts a prior expectation that the generation of the most recent common
    ancestor (MRCA) between extant hereditary stratigraphic columns becomes
    exponentialy less likely with increasing antiquity.

    This prior calculates the exact, discrete geometric distribution of time to
    MRCA expected under the Wright-Fisher model [1].

    .. [1] Alison Etheridge (2006). Course 11 - Evolution in Fluctuating
    Populations. In Mathematical Statistical Physics (pp. 489-545). Elsevier.
    https://doi.org/10.1016/S0924-8099(06)80048-X

    Notes
    -----
    A static factory function to init an instance with a growth factor
    calculated from contextual information like population size and the number
    of generations elapsed since genesis should be made available in the
    future.

    See Also
    --------
    GeometricPrior
        A continuous approximation of this prior.
    """

    _growth_factor: float

    def __init__(self: "GeometricPrior", growth_factor: float):
        self._growth_factor = growth_factor

    def CalcIntervalProbabilityProxy(
        self: "GeometricPrior", begin_rank: int, end_rank: int
    ) -> float:
        """Characterize the prior probability of the MRCA generation falling
        within an interval range.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        float
            The proxy statistic, proportional to the true estimated interval
            probability of the MRCA value by a fixed (but unspecified) constant
            proportion.
        """
        f = self._growth_factor
        return np.logspace(
            begin_rank,
            end_rank,
            base=f,
            endpoint=False,
            num=end_rank - begin_rank,
        ).sum()

    def CalcIntervalConditionedMean(
        self: "GeometricPrior", begin_rank: int, end_rank: int
    ) -> float:
        """Calculate the centriod of prior probability mass within an interval
        of possible MRCA generations.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        float
            The prior expected generation of MRCA conditioned on the assumption
            that the MRCA falls within the given interval.
        """
        f = self._growth_factor
        return np.average(
            np.arange(begin_rank, end_rank),
            weights=np.logspace(
                begin_rank,
                end_rank,
                base=f,
                endpoint=False,
                num=end_rank - begin_rank,
            ),
        )

    def SampleIntervalConditionedValue(
        self: "GeometricPrior", begin_rank: int, end_rank: int
    ) -> int:
        """Sample a generation of the MRCA conditioned on the assumption that
        the MRCA falls within the given interval.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        int
            A sampled generation of the MRCA, conditioned on the assumption that
            the MRCA falls within the given interval.
        """
        raise NotImplementedError()



import os

import polars as pl

from hstrat._auxiliary_lib import (
    alifestd_is_chronologically_ordered,
    alifestd_try_add_ancestor_list_col,
    alifestd_validate,
)
from hstrat.dataframe import (
    surface_postprocess_trie,
    surface_unpack_reconstruct,
)
from hstrat.phylogenetic_inference.tree.trie_postprocess import (
    AssignOriginTimeNodeRankTriePostprocessor,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_smoke():
    df = pl.read_csv(f"{assets_path}/packed.csv")
    raw = surface_unpack_reconstruct(df)
    res = surface_postprocess_trie(
        raw,
        trie_postprocessor=AssignOriginTimeNodeRankTriePostprocessor(),
    )
    assert "origin_time" in res.columns
    assert len(res) <= len(raw)
    assert alifestd_validate(
        alifestd_try_add_ancestor_list_col(res.to_pandas()),
    )
    assert alifestd_is_chronologically_ordered(res.to_pandas())



"""Data structures to annotate genomes with."""

from . import stratum_ordered_stores
from .._auxiliary_lib import lazy_attach

__getattr__, __dir__, __all__ = lazy_attach(
    __name__,
    submodules=["stratum_ordered_stores"],
    submod_attrs={
        "stratum_ordered_stores": stratum_ordered_stores.__all__,
        "_HereditaryStratigraphicColumn": ["HereditaryStratigraphicColumn"],
        "_HereditaryStratum": ["HereditaryStratum"],
        "_HereditaryStratigraphicColumnBundle": [
            "HereditaryStratigraphicColumnBundle",
        ],
    },
    should_launder=[
        "HereditaryStratigraphicColumn",
        "HereditaryStratum",
        "HereditaryStratigraphicColumnBundle",
    ].__contains__,
)
del lazy_attach



from hstrat._auxiliary_lib import GetAttrLaunderShim


def test_GetAttrLaunderShim_dummy1():
    pass


def test_GetAttrLaunderShim_dummy2():
    pass


def test_GetAttrLaunderShim():
    shim = GetAttrLaunderShim(
        globals().__getitem__,  # current module
        ["test_GetAttrLaunderShim_dummy1"].__contains__,
        "foobar",
    )
    a = shim("test_GetAttrLaunderShim_dummy1")
    assert a.__module__ == "foobar"

    b = shim("test_GetAttrLaunderShim_dummy2")
    assert b.__module__ == test_GetAttrLaunderShim.__module__



"""Maintain constant size complexity with recency-proportional strata spacing.

The exactly space-filling MRCA-recency-proportional resolution policy
imposes an O(1) limit on the number of retained strata and guarantees that
retained strata will be exponentially distributed with respect to ranks elapsed
since their deposit. MRCA rank estimate uncertainty scales in the worst case
scales as O(n) with respect to the greater number of strata deposited on either
column. However, with respect to estimating the rank of the MRCA when lineages
diverged any fixed number of generations ago, uncertainty scales as O(log(n))
(TODO check this).

Under the MRCA-recency-proportional resolution policy, the number of strata
retained (i.e., space complexity) scales as O(1) with respect to the number of
strata deposited.

Suppose k is specified as the policy's target space utilization. All strata
will be retained until an upper a hard upper limit of 4k + 2 is reached. Then,
the number of strata is maintained exactly at this hard upper limit perpetually.
MRCA rank estimate uncertainty is maintained less than or equal to
s * (1 - n^(-1/k)) where n is the number of strata deposited and s is the true
number of ranks deposited since the MRCA.

See Also
--------
geom_seq_nth_root_algo:
    For a predicate retention policy that achieves the same guarantees for
    resolution and space complexity but fluctuates in size below an upper
    size bound instead of remaining exactly at the size bound. Likely faster
    than tapered_geom_seq_nth_root_algo for most operations.
"""

from ...._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



import typing

import numpy as np
import pandas as pd

from ..perfect_tracking import GarbageCollectingPhyloTracker
from ._evolve_fitness_trait_population_ import (
    _apply_island_migrations,
    _apply_mutation,
    _apply_niche_invasions,
    _get_island_id,
    _get_niche_id,
    _select_parents,
)


def evolve_fitness_trait_population(
    population_size: int = 1024,
    iter_epochs: bool = False,
    num_islands: int = 4,
    num_niches: int = 4,
    num_generations: int = 100,
    tournament_size: int = 4,
    p_island_migration: float = 1e-3,
    p_niche_invasion: float = 1e-4,
    mut_distn: typing.Callable = np.random.standard_normal,
    share_common_ancestor: bool = True,
    progress_wrap: typing.Callable = lambda x: x,
    tracker_buffer_size: typing.Optional[int] = None,
) -> typing.Union[pd.DataFrame, typing.Iterator[pd.DataFrame]]:
    """Run simple evolutionary simulation to generate sample phylogeny.

    Organisms are simulated with a genetic trait, which directly corresponds to
    fitness. Mutations to this genetic trait are performed every generation.
    Tournament selection determines the composition of synchronous generations.
    One tournament is held for each slot within the upcoming generation, and
    the organism within a tournament with the highest genetic trait value is
    selected as the asexual parent of the organism to be placed in that slot.

    A simple island model allows incorporation of spatial structure into the
    simulation. Islands are arranged in a ring topology. Population size is
    evenly distributed across islands. In selection tournaments, organisms only
    compete against other organisms from the same island.

    Organisms migrate to either neighboring island with fixed per-organism per-
    generation probability. To maintain exactly fixed per-island population
    size, a copy of the migrating organism remains on the original island and
    an additional copy is placed on the target island, overwriting an
    inhabitant of that island. Care is taken such that the net number of
    migrations per generation corresponds to a binomial distribution with `n =
    population_size` and `p = p_island_migration`. A single organism may
    yield up to two migrant copies within a single generation (one migrating to
    each island). Migrant copies cannot yield further migrant copies within the
    same generation.

    A simple niche model allows incorporation of ecological dynamics into the
    simulation. A fixed number of niches are instantiated. In selection
    tournaments, organisms only compete against other organisms in the same
    niche. Islands' population capacity is evenly split between niches. That is,
    each niche hosts the same number of members within an island.

    Organisms switch to another niche ("invade" that niche) with fixed per-
    organism per-generation probability. The switched-to niche is determined
    randomly with all other niches having uniform probability of being chosen.
    To maintain exactly fixed niche population size, a copy of the invading
    organism remains in the original niche and an additional copy is created as
    a member of the invaded niche, overriding an existing member of the invaded
    niche. Care is taken such that the net number of niche invasions per
    generation corresponds to a binomial distribution with `n =
    population_size` and `p = p_niche_invasion`. A single organism may yield up
    to one invading copy per non-self niche available.

    Parameters
    ----------
    population_size : int, default 1024
        Number of organisms within each generational cohort.
    iter_epochs : bool, default False
        Should an iterator that yields phylogenies at successive time points
        be returned instead of a single final phylogeny?
    num_islands : int, default 4
        Must evenly divide `population_size`.
    num_niches : int, default 4
        Must evenly divide `population_size`.
    num_generations : int, default 100
        Number of generations to simulate.

        If `iter_epochs` is True, the number of generations to simulate within
        an epoch.
    tournament_size : int, default 4
        Number of organisms sampled to compete for each population slot.

        Higher `tournament_size` increases selection pressure. If set to 1,
        selection constitutes neutral drift.
    p_island_migration : float, default 1e-3
        Per-organism, per-generation probability of copying an organism into a
        neighboring island.
    p_niche_invasion : float, default 1e-4
        Per-organism, per-generation probability of copying an organism into a
        neighboring island.
    mut_distn : typing.Callable, default numpy.random.standard_normal
        Callable to generate mutation deltas applied to each organism at each
        generation.

        Must take int value `size` as sole kwarg. Must return a `numpy.array`
        of float with length `size`.
    share_common_ancestor : bool, default True
        Should a dummy common ancestor be inserted as the first entry in
        the phylogenetic record?

        If True, all initial population members will be recorded as
        children of this dummy ancestor. If False, all initial
        population members will be recorded as having no parent.

        The dummy ancestor's trait value is recorded as NaN and population
        loc is recorded as 0.
    progress_wrap : Callable, optional
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.
    tracker_buffer_size : int, optional
        How many rows back should intermittent garbage collection on
        phylogenetic tracker reach?

        Adjustments to this parameter only affect performance, not tracking semantics. If None, a default value will be used.
    Returns
    -------
    pandas.DataFrame or iterator of pandas.DataFrame
        Full phylogenetic record of extant organisms in alife standard format
        either (1) at end of evolutionary simulation after `num_generations` or
        (2) at `num_generations` intervals as an infinite iterator.

        Return mode depends on `iter_epochs` parameter.

    Notes
    -----
    The product of `num_islands` and `num_niches` must divide
    `population_size`.
    """

    island_size = population_size // num_islands
    island_niche_size = population_size // (num_islands * num_niches)
    assert island_niche_size * num_islands * num_niches == population_size
    assert tournament_size <= island_niche_size

    pop_arr = np.zeros(population_size, dtype=np.single)
    pop_tracker = GarbageCollectingPhyloTracker(
        pop_arr,
        share_common_ancestor=share_common_ancestor,
        working_buffer_size=tracker_buffer_size,
    )

    def epoch_iterator(
        pop_arr: np.ndarray,
        pop_tracker: GarbageCollectingPhyloTracker,
    ) -> typing.Iterator[pd.DataFrame]:
        """Infinite iterator that yields cumulative phylogeny at evenly-spaced
        generational epochs."""

        while True:
            for __ in progress_wrap(range(num_generations)):
                _apply_island_migrations(
                    pop_arr,
                    pop_tracker,
                    num_niches=num_niches,
                    island_size=island_size,
                    island_niche_size=island_niche_size,
                    p_island_migration=p_island_migration,
                )
                _apply_niche_invasions(
                    pop_arr,
                    pop_tracker,
                    num_islands=num_islands,
                    num_niches=num_niches,
                    island_size=island_size,
                    island_niche_size=island_niche_size,
                    p_niche_invasion=p_niche_invasion,
                )
                _apply_mutation(pop_arr, mut_distn)
                parent_locs = _select_parents(
                    island_niche_size=island_niche_size,
                    tournament_size=tournament_size,
                    pop_arr=pop_arr,
                )

                pop_arr = pop_arr[parent_locs]
                pop_tracker.ElapseGeneration(
                    parent_locs,
                    pop_arr,
                )

            yield pop_tracker.CompilePhylogeny(
                loc_transforms={
                    "island": lambda loc: _get_island_id(loc, island_size),
                    "niche": lambda loc: _get_niche_id(
                        loc, island_niche_size, num_niches
                    ),
                },
                progress_wrap=progress_wrap,
            )

    if iter_epochs:
        return epoch_iterator(pop_arr, pop_tracker)
    else:
        return next(epoch_iterator(pop_arr, pop_tracker))



import anytree


def anytree_has_sibling(node: anytree.node) -> bool:
    """Check whether the given node has at least one sibling in the Anytree
    tree.

    Parameters
    ----------
    node : anytree.node
        The node to check for siblings.

    Returns
    -------
    bool
        True if the node has at least one sibling in the tree, False otherwise.
    """
    return node.parent is not None and len(node.parent.children) > 1



import typing

from packaging import version
import pandas as pd

from ..._auxiliary_lib import (
    HereditaryStratigraphicArtifact,
    get_hstrat_version,
)
from ..priors._ArbitraryPrior import ArbitraryPrior
from ._build_tree_trie import build_tree_trie


def build_tree(
    population: typing.Sequence[HereditaryStratigraphicArtifact],
    version_pin: str,
    taxon_labels: typing.Optional[typing.Iterable] = None,
    force_common_ancestry: bool = False,
    progress_wrap: typing.Callable = lambda x: x,
) -> pd.DataFrame:
    """Estimate the phylogenetic history among hereditary stratigraphic
    columns.

    This function provides a generic interface that directs to an underlying
    implementation, chosen for flexibility and robustness across reconstruction
    scenarios. The underlying implementation may be revised in future releases
    of the software.

    The current implementation delegates to `build_tree_trie`.

    Parameters
    ----------
    population: Sequence[HereditaryStratigraphicArtifact]
        Hereditary stratigraphic columns corresponding to extant population members.

        Each member of population will correspond to a unique leaf node in the
        reconstructed tree.
    version_pin: str
        How should calls to this function resolve in future hstrat releases?

        To prevent future releases from silently substituting the underlying
        reconstruction algorithm, hardcode the current version of hstrat here.
        To automatically track any future library updates, pass
        `hstrat.__version__` here.

        Some effort will be made to maintain historical implementations to
        support prior version pins. However, indefinite support is not
        guaranteed for hard version pins; old version pins may eventually raise
        `DeprecationWarning` or `ValueError`. Where reasonable, consider
        directly calling an implementing tree building method instead.
    taxon_labels: Optional[Iterable], optional
        How should leaf nodes representing extant hereditary stratigraphic
        columns be named?

        Label order should correspond to the order of corresponding hereditary
        stratigraphic columns within `population`. If None, taxons will be
        named according to their numerical index.
    force_common_ancestry: bool, optional
        How should columns that definitively share no common ancestry be
        handled?

        If set to True, treat columns with no common ancestry as if they
        shared a common ancestor immediately before the genesis of the
        lineages. If set to False, columns within `population` that
        definitively do not share common ancestry will raise a ValueError.
    progress_wrap : Callable, optional
        Pass tqdm or equivalent to display progress bars.

    Returns
    -------
    pd.DataFrame
        The reconstructed phylogenetic tree in alife standard format.

    Raises
    ------
    ValueError
        If the specified `version_pin` is higher than the current version of
        hstrat.
    """

    if version.parse(version_pin) > version.parse(get_hstrat_version()):
        raise ValueError(f"unsupported verison {version_pin}")

    return build_tree_trie(
        population,
        taxon_labels,
        force_common_ancestry=force_common_ancestry,
        progress_wrap=progress_wrap,
        bias_adjustment=ArbitraryPrior(),
    )



import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._estimate_ranks_since_mrca_with import estimate_ranks_since_mrca_with


def ballpark_ranks_since_mrca_with(
    focal: HereditaryStratigraphicArtifact,
    other: HereditaryStratigraphicArtifact,
) -> typing.Optional[float]:
    """Calculate a fast, rough estimate of generations elapsed since MRCA with
    other.

    See `estimate_ranks_since_mrca_with` for details.
    """
    return estimate_ranks_since_mrca_with(
        focal,
        other,
        estimator="maximum_likelihood",
        prior="arbitrary",
    )



import typing

import pandas as pd
import polars as pl

from ...._auxiliary_lib import (
    AnyTreeFastPreOrderIter,
    anytree_iterative_deepcopy,
)
from .._impl import TrieInnerNode
from ._detail import TriePostprocessorBase


def _call_anytree(
    ftor: "AssignOriginTimeNodeRankTriePostprocessor",
    trie: TrieInnerNode,
    progress_wrap: typing.Callable,
) -> TrieInnerNode:
    for node in progress_wrap(AnyTreeFastPreOrderIter(trie)):
        t0 = getattr(node, ftor._t0) if isinstance(ftor._t0, str) else ftor._t0
        setattr(node, ftor._assigned_property, node.rank - t0)

    return trie


def _call_pandas(
    ftor: "AssignOriginTimeNodeRankTriePostprocessor",
    trie: pd.DataFrame,
) -> TrieInnerNode:
    t0 = trie[ftor._t0] if isinstance(ftor._t0, str) else ftor._t0
    trie[ftor._assigned_property] = trie["rank"] - t0
    return trie


def _call_polars(
    ftor: "AssignOriginTimeNodeRankTriePostprocessor",
    trie: pl.DataFrame,
) -> TrieInnerNode:
    t0 = pl.col(ftor._t0) if isinstance(ftor._t0, str) else pl.lit(ftor._t0)
    return trie.with_columns(
        (pl.col("rank") - t0).alias(ftor._assigned_property),
    )


class AssignOriginTimeNodeRankTriePostprocessor(TriePostprocessorBase):
    """Functor to assign trie nodes' rank as their the origin time."""

    _assigned_property: str
    _t0: int

    def __init__(
        self: "AssignOriginTimeNodeRankTriePostprocessor",
        assigned_property: str = "origin_time",
        *,
        t0: typing.Union[int, str] = 0,
    ) -> None:
        """Initialize functor instance.

        Parameters
        ----------
        assigned_property : str, default "origin_time"
            The property name for the assigned origin time.
        t0 : int or str, default 0
            The property name or constant value for the origin time offset.
        """
        self._assigned_property = assigned_property
        self._t0 = t0

    def __call__(
        self: "AssignOriginTimeNodeRankTriePostprocessor",
        trie: TrieInnerNode,
        p_differentia_collision: float,
        mutate: bool = False,
        progress_wrap: typing.Callable = lambda x: x,
    ) -> TrieInnerNode:
        """Assign origin times to trie nodes.

        Parameters
        ----------
        trie : TrieInnerNode
            The input trie to be postprocessed.
        p_differentia_collision : float
            Probability of a randomly-generated differentia matching an
            existing differentia.
            Not used in the current implementation.
        mutate : bool, default False
            Are side effects on the input argument `trie` allowed?
        progress_wrap : typing.Callable, optional
            Pass tqdm or equivalent to report progress.

        Returns
        -------
        TrieInnerNode
            The postprocessed trie with assigned origin times.
        """
        if isinstance(trie, TrieInnerNode):
            if not mutate:
                trie = anytree_iterative_deepcopy(
                    trie, progress_wrap=progress_wrap
                )
            return _call_anytree(
                self,
                trie,
                progress_wrap=progress_wrap,
            )
        elif isinstance(trie, pd.DataFrame):
            if not mutate:
                trie = trie.copy()
            return _call_pandas(
                self,
                trie,
            )  # no progress wrap
        elif isinstance(trie, pl.DataFrame):
            if not mutate:
                trie = trie.clone()
            return _call_polars(
                self,
                trie,
            )  # no progress wrap
        else:
            raise TypeError  # pragma: no cover



from base64 import b64encode

import typing_extensions


def stringify_packed_differentia_bytes(
    buffer: typing_extensions.Buffer,
) -> str:
    """Convert packed differentia bytes to a string representation."""
    encoded_bytes = b64encode(buffer)
    encoded_str = encoded_bytes.decode()
    return encoded_str



"""Functions that compare two columns.

Provides the foundation for phylogenetic inference tools.
"""

from hstrat._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



import sys


def is_in_coverage_run() -> bool:
    """Check whether test is run with coverage."""
    gettrace = getattr(sys, "gettrace", None)

    if gettrace is None:  # pragma: no cover
        return False
    else:
        gettrace_result = gettrace()

    try:
        from coverage.pytracer import PyTracer
        from coverage.tracer import CTracer

        if isinstance(gettrace_result, (CTracer, PyTracer)):
            return True
    except ImportError:  # pragma: no cover
        pass

    return False  # pragma: no cover



import pytest

from hstrat._auxiliary_lib import deep_listify


def test_deep_listify_not_iterable():
    result = deep_listify(1)
    assert result == 1

    result = deep_listify(None)
    assert result is None

    result = deep_listify(False)
    assert result is False


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_deep_listify_empty_list(wrap):
    result = deep_listify(wrap([]))
    assert result == []


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_deep_listify_empty_list_nested(wrap):
    result = deep_listify(wrap([wrap([])]))
    assert result == [[]]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_deep_listify_flat_list(wrap):
    result = deep_listify(wrap([1, 2, 3]))
    assert result == [1, 2, 3]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_deep_listify_nested_list(wrap):
    result = deep_listify(wrap([[1, 2], [3, 4]]))
    assert result == [[1, 2], [3, 4]]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_deep_listify_nested_tuple(wrap):
    result = deep_listify(wrap([(1, 2), (3, 4)]))
    assert result == [[1, 2], [3, 4]]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_deep_listify_mixed_nested(wrap):
    result = deep_listify(wrap([1, wrap([2, wrap((3, 4))]), 5]))
    assert result == [1, [2, [3, 4]], 5]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_deep_listify_iterator(wrap):
    result = deep_listify(wrap(iter([1, 2, 3])))
    assert result == [1, 2, 3]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_deep_listify_generator(wrap):
    result = deep_listify(wrap((x for x in range(3))))
    assert result == [0, 1, 2]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_deep_listify_singleton_list_nested(wrap):
    result = deep_listify(wrap([wrap(["asdf"])]))
    assert result == [["asdf"]]



import typing
import warnings

from .._detail import PolicySpecBase


class PolicySpec(PolicySpecBase):
    """Contains all policy parameters, if any."""

    _size_curb: int

    def __init__(
        self: "PolicySpec",
        size_curb: int,
    ) -> None:
        """Construct the policy spec.

        Parameters
        ----------
        size_curb : int, optional
            Fixed, hard-limited stratum storage capacity to make best use of.

            Inclusive. Only size caps of eight strata and above are guaranteed
            respected. (Initialization of smaller size caps will warn.)
        """
        self._size_curb = size_curb
        if size_curb < 8:
            warnings.warn(
                f"Size curb set to {size_curb}, "
                "the minimum guaranteeable size limit is 8.",
            )

    def __eq__(self: "PolicySpec", other: typing.Any) -> bool:
        return isinstance(other, self.__class__) and (self._size_curb,) == (
            other._size_curb,
        )

    def __repr__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoIdentifier()
        }.{
            PolicySpec.__qualname__
        }(size_curb={
            self._size_curb
        })"""

    def __str__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoTitle()
        } (size_curb: {
            self._size_curb
        })"""

    def GetEvalCtor(self: "PolicySpec") -> str:
        return f"hstrat.{self!r}"

    def GetSizeCurb(self: "PolicySpec") -> int:
        return self._size_curb

    @staticmethod
    def GetAlgoIdentifier() -> str:
        """Get programatic name for underlying retention algorithm."""
        return __package__.split(".")[-1]

    @staticmethod
    def GetAlgoTitle() -> str:
        """Get human-readable name for underlying retention algorithm."""
        return "Curbed Recency-proportional Resolution Stratum Retention Algorithm"



import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._estimate_rank_of_mrca_between import estimate_rank_of_mrca_between


def ballpark_rank_of_mrca_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
) -> typing.Optional[float]:
    """Calculate a fast, rough estimate of the rank of the MRCA beteen first
    and second.

    See `estimate_rank_of_mrca_between` for details.
    """
    return estimate_rank_of_mrca_between(
        first,
        second,
        estimator="maximum_likelihood",
        prior="arbitrary",
    )



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec
from .._impl import calc_provided_uncertainty


class CalcRankAtColumnIndex:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcRankAtColumnIndex",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcRankAtColumnIndex", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcRankAtColumnIndex",
        policy: PolicyCouplerBase,
        index: int,
        num_strata_deposited: typing.Optional[int],
    ) -> int:
        """After n strata have been deposited, what will the rank of the
        stratum at column index k be?

        Enables a HereditaryStratigraphicColumn using this predicate to
        optimize away storage of rank annotations on strata. Takes into the
        account the possibility for in-progress stratum depositions that haven't
        been reflected in num_strata_deposited.
        """
        spec = policy.GetSpec()
        resolution = spec.GetRecencyProportionalResolution()

        # calculate the interval between retained strata we're starting out with
        # -1 due to *lack* of an in-progress deposition
        provided_uncertainty = calc_provided_uncertainty(
            resolution,
            num_strata_deposited - 1,
        )
        # we could just take a simple recursive approach like this
        #
        #   if index == 0: return 0
        #   else: return provided_uncertainty + self.CalcRankAtColumnIndex(
        #       index - 1,
        #       num_strata_deposited - provided_uncertainty,
        #   )
        #
        # but as an optimization, we should take as many steps as possible at
        # the current uncertainty interval size stage before recursing down to
        # the next half-as-small uncertainty interval size

        # calculate the greatest rank at which that interval is viable
        # i.e., where max_uncertainty == provided_uncertainty... we must solve
        #
        #   provided_uncertainty = least_num_strata // (resolution + 1)
        #
        # and
        #
        #   least_num_strata = num_strata_deposited - greatest_viable_rank

        # -1 due to *lack* of an in-progress deposition
        greatest_viable_rank = (
            num_strata_deposited - 1 - provided_uncertainty * (resolution + 1)
        )

        # calculate how many steps at provided_uncertainty interval we can take
        # +provided_uncertainty because greatest viable rank is the *last*
        # position we can take a step in our interval from
        num_interval_steps = (
            greatest_viable_rank + provided_uncertainty
        ) // provided_uncertainty

        if index <= num_interval_steps or provided_uncertainty == 1:
            # we can reach index within the current provided_uncertainty stage
            # note: must always enter base case when provided uncertainty is 1
            # or infinite recursion will result
            return index * provided_uncertainty
        else:
            # take as many index steps as possible at current uncertainty stage
            # and add number of ranks accrued at subsequent uncertainty stages
            num_ranks_traversed = num_interval_steps * provided_uncertainty
            return num_ranks_traversed + self(
                policy=policy,
                index=index - num_interval_steps,
                num_strata_deposited=num_strata_deposited
                - num_ranks_traversed,
            )



import itertools as it

import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
def test_GetNthCommonRankWith(wrap):
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=hstrat.nominal_resolution_algo.Policy(),
    )
    c3 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
    )
    for __ in range(42):
        c1.DepositStratum()

    for __ in range(9):
        c2.DepositStratum()

    for __ in range(100):
        c3.DepositStratum()

    for col in c1, c2, c3:
        for i in range(col.GetNumStrataRetained()):
            assert col.GetRankAtColumnIndex(
                i
            ) == hstrat.get_nth_common_rank_between(wrap(col), wrap(col), i)
        assert (
            hstrat.get_nth_common_rank_between(
                wrap(col), wrap(col), col.GetNumStrataRetained()
            )
            is None
        )

    for x1, x2 in it.combinations([c1, c2, c3], 2):
        assert hstrat.get_nth_common_rank_between(wrap(x1), wrap(x2), 0) == 0

    assert (
        hstrat.get_nth_common_rank_between(wrap(c1), wrap(c2), 1)
        == c2.GetNumStrataDeposited() - 1
    )
    assert (
        hstrat.get_nth_common_rank_between(wrap(c2), wrap(c1), 1)
        == c2.GetNumStrataDeposited() - 1
    )

    assert hstrat.get_nth_common_rank_between(wrap(c1), wrap(c2), 2) is None
    assert hstrat.get_nth_common_rank_between(wrap(c2), wrap(c1), 2) is None

    assert hstrat.get_nth_common_rank_between(wrap(c1), wrap(c3), 1) == 2
    assert hstrat.get_nth_common_rank_between(wrap(c3), wrap(c1), 1) == 2

    assert hstrat.get_nth_common_rank_between(wrap(c1), wrap(c3), 2) == 4
    assert hstrat.get_nth_common_rank_between(wrap(c3), wrap(c1), 2) == 4


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
def test_artifact_types_equiv(differentia_width, policy):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        for n in range(10):
            assert hstrat.get_nth_common_rank_between(
                hstrat.col_to_specimen(a),
                hstrat.col_to_specimen(b),
                n,
            ) == hstrat.get_nth_common_rank_between(a, b, n)



import pandas as pd


def alifestd_make_ancestor_id_col(
    ids: pd.Series, ancestor_lists: pd.Series
) -> pd.Series:
    """Translate ancestor ids from a column of singleton `ancestor_list`s into a
    pure-integer series representation.

    Each organism must have one or zero ancestors (i.e., asexualasexual data).
    In the returned series, ancestor id will be assigned to own id for no-
    ancestor organisms.
    """
    ancestor_ids = (
        ancestor_lists.astype("str")
        .str.lower()
        .replace("[none]", "[-1]")
        .replace("[]", "[-1]")
        .str.strip("[]")
        .astype(int)
    )

    root_filter = ancestor_ids == -1
    ancestor_ids[root_filter] = ids[root_filter]

    return ancestor_ids



import os

import pkg_resources


# Adapted from https://stackoverflow.com/a/3787989
def check_testing_requirements() -> None:
    """Are all testing requirements available?

    If dependencies are missing, will raise pkg_resources.DistributionNotFound
    or pkg_resources.VersionConflict warning.
    """

    testing_requirements_path = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "..",
            "requirements-dev",
            "requirements-testing.txt",
        )
    )
    with open(testing_requirements_path, "r") as testing_requirements:
        # adapted from https://stackoverflow.com/a/16298328
        pkg_resources.require(testing_requirements)



import typing

from .._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)
from ._calc_ranks_since_last_retained_commonality_with import (
    calc_ranks_since_last_retained_commonality_with,
)


def calc_definitive_min_ranks_since_last_retained_commonality_with(
    focal: HereditaryStratigraphicArtifact,
    other: HereditaryStratigraphicArtifact,
) -> typing.Optional[int]:
    """Determine hard, inclusive lower bound on generations since MRCA.

    At least, how many depositions have elapsed along the focal column's line
    of descent since the last matching strata at the same rank between focal
    and other?

    Returns None if no common ancestor between focal and other can be
    resolved with absolute confidence.
    """
    confidence_level = 0.49
    assert (
        focal.GetStratumDifferentiaBitWidth()
        == other.GetStratumDifferentiaBitWidth()
    )
    assert (
        calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            focal,
            other,
            significance_level=1.0 - confidence_level,
        )
        == 1
    )
    return calc_ranks_since_last_retained_commonality_with(
        focal,
        other,
        confidence_level=confidence_level,
    )



import pandas as pd

from ._alifestd_assign_contiguous_ids import alifestd_assign_contiguous_ids
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_to_working_format(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Re-encode phylogeny_df to facilitate efficient analysis and
    transformation operations.

    The returned phylogeny dataframe will
    * be topologically sorted (i.e., organisms appear after all ancestors),
    * have contiguous ids (i.e., organisms' ids correspond to row number),
    * contain an integer datatype `ancestor_id` column if the phylogeny is
    asexual (i.e., a more performant representation of `ancestor_list`).

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    if not alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df = alifestd_assign_contiguous_ids(
            phylogeny_df, mutate=True
        )

    return phylogeny_df



import contexttimer as ctt
import more_itertools as mit
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import is_base64


@pytest.mark.parametrize(
    "differentia_bit_width",
    [
        1,
        8,
        16,
        32,
        64,
    ],
)
@pytest.mark.parametrize(
    "num_strata",
    [
        0,
        1,
        2,
        7,
        8,
        15,
        27,
        32,
    ],
)
def test_pack_differentiae_str(differentia_bit_width, num_strata):
    original_strata = [
        hstrat.HereditaryStratum(
            differentia_bit_width=differentia_bit_width,
            deposition_rank=i,
        )
        for i in range(num_strata)
    ]
    packed_differentiae = hstrat.pack_differentiae_str(
        original_strata,
        differentia_bit_width,
    )
    assert is_base64(packed_differentiae)

    unpacked_differentiae = hstrat.unpack_differentiae_str(
        packed_differentiae,
        differentia_bit_width,
        differentiae_byte_bit_order="big",
    )
    if num_strata >= 20:
        assert unpacked_differentiae != hstrat.unpack_differentiae_str(
            packed_differentiae,
            differentia_bit_width,
            differentiae_byte_bit_order="little",
        )

    for stratum, differentia in mit.zip_equal(
        original_strata,
        unpacked_differentiae,
    ):
        assert stratum.GetDifferentia() == differentia


@pytest.mark.parametrize(
    "differentia_bit_width",
    [
        1,
        8,
        16,
        32,
        64,
    ],
)
def test_benchmark(differentia_bit_width):
    num_strata = 10000
    original_strata = [
        hstrat.HereditaryStratum(
            differentia_bit_width=differentia_bit_width,
            deposition_rank=i,
        )
        for i in range(num_strata)
    ]

    packed_differentiae = hstrat.pack_differentiae_str(
        original_strata,
        differentia_bit_width,
    )

    with ctt.Timer(factor=1000) as timer:
        hstrat.unpack_differentiae_str(
            packed_differentiae,
            differentia_bit_width,
        )

    print()
    print(
        f"differentia_bit_width={differentia_bit_width}",
        f"num_strata={num_strata}",
        f"timer.elapsed={timer.elapsed}",
    )



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_make_empty,
    alifestd_mark_origin_time_delta_asexual,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_empty():
    mt = alifestd_make_empty()
    mt["origin_time"] = None
    res = alifestd_mark_origin_time_delta_asexual(mt)
    assert "ancestor_origin_time" in res
    assert "origin_time_delta" in res
    assert len(res) == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "origin_time": [0, 10, 30],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_origin_time_delta_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    assert result_df.loc[0, "ancestor_origin_time"] == 0
    assert result_df.loc[1, "ancestor_origin_time"] == 0
    assert result_df.loc[2, "ancestor_origin_time"] == 10
    assert result_df.loc[0, "origin_time_delta"] == 0
    assert result_df.loc[1, "origin_time_delta"] == 10
    assert result_df.loc[2, "origin_time_delta"] == 20

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3],
            "ancestor_list": ["[0]", "[None]", "[0]", "[1]"],
            "origin_time": [20, 10, 30, 45],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_origin_time_delta_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[1, "ancestor_origin_time"] == 10
    assert result_df.loc[0, "ancestor_origin_time"] == 10
    assert result_df.loc[2, "ancestor_origin_time"] == 10
    assert result_df.loc[3, "ancestor_origin_time"] == 20

    assert result_df.loc[1, "origin_time_delta"] == 10
    assert result_df.loc[0, "origin_time_delta"] == 0
    assert result_df.loc[2, "origin_time_delta"] == 20
    assert result_df.loc[3, "origin_time_delta"] == 25

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple3(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3],
            "ancestor_list": ["[None]", "[None]", "[0]", "[1]"],
            "origin_time": [20, 10, 30, 45],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_origin_time_delta_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[1, "ancestor_origin_time"] == 20
    assert result_df.loc[0, "ancestor_origin_time"] == 10
    assert result_df.loc[2, "ancestor_origin_time"] == 10
    assert result_df.loc[3, "ancestor_origin_time"] == 20
    assert result_df.loc[1, "origin_time_delta"] == 0
    assert result_df.loc[0, "origin_time_delta"] == 0
    assert result_df.loc[2, "origin_time_delta"] == 20
    assert result_df.loc[3, "origin_time_delta"] == 25

    if not mutate:
        assert original_df.equals(phylogeny_df)



import typing


def flat_len(maybe_iterable: typing.Iterable) -> int:
    """Count the total number of elements in an iterable, recursing to count
    elements in any nested iterables.

    Examples
    --------
    >>> flat_len(range(4))
    4

    >>> flat_len(4)
    1

    >>> flat_len([(1, 2), 3, 4])
    4

    >>> flat_len([1, [2, (3, 4)], 5])
    [1, [2, [3, 4]], 5]

    >>> flat_len(iter([1, range(2), 3]))
    4
    """
    if isinstance(maybe_iterable, str):
        return len(maybe_iterable)
    try:
        return sum(map(flat_len, maybe_iterable))
    except TypeError:
        return 1



import typing

from interval_search import binary_search

from ...genome_instrumentation import HereditaryStratigraphicColumn
from .._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)
from ._calc_rank_of_first_retained_disparity_between_generic import (
    calc_rank_of_first_retained_disparity_between_generic,
)


def calc_rank_of_first_retained_disparity_between_bsearch(
    first: HereditaryStratigraphicColumn,
    second: HereditaryStratigraphicColumn,
    confidence_level: float,
) -> typing.Optional[int]:
    """Find first mismatching strata between columns.

    Implementation detail. Provides optimized implementation for special
    case where both self and second use the perfect resolution stratum
    retention policy.
    """
    # both must have (effectively) used the perfect resolution policy
    assert not first.HasDiscardedStrata() and not second.HasDiscardedStrata()

    lower_bound = 0
    upper_bound = min(
        [
            first.GetNumStrataDeposited() - 1,
            second.GetNumStrataDeposited() - 1,
        ]
    )
    assert lower_bound <= upper_bound

    def differentia_at(
        which: HereditaryStratigraphicColumn,
        idx: int,
    ) -> int:
        return which.GetStratumAtColumnIndex(idx).GetDifferentia()

    def predicate(idx: int) -> bool:
        return differentia_at(first, idx) != differentia_at(second, idx)

    first_disparite_idx = binary_search(
        predicate,
        lower_bound,
        upper_bound,
    )

    if first_disparite_idx is not None:
        # disparate strata found
        assert (
            first.GetStratumDifferentiaBitWidth()
            == second.GetStratumDifferentiaBitWidth()
        )
        collision_plausibility_threshold = (
            calc_min_implausible_spurious_consecutive_differentia_collisions_between(
                first,
                second,
                significance_level=1.0 - confidence_level,
            )
            - 1
        )
        # discount collision_implausibility_threshold - 1 common
        # ranks due to potential spurious differentia collisions;
        # if not enough common ranks are available we still know
        # *definitively* that a disparity occured (because we
        # observed disparite strata at the same rank); so, make the
        # conservative assumption that the disparity occured as far
        # back as possible (rank 0)
        spurious_collision_corrected_idx = max(
            first_disparite_idx - collision_plausibility_threshold,
            0,
        )
        first_disparite_rank = first.GetRankAtColumnIndex(
            spurious_collision_corrected_idx
        )
        assert first_disparite_rank >= 0
        return first_disparite_rank
    else:
        # no disparate strata found
        # fall back to calc_rank_of_first_retained_disparity_between_generic to
        # handle proper bookkeeping in this case while skipping most of the
        # search
        return calc_rank_of_first_retained_disparity_between_generic(
            first,
            second,
            first_start_idx=upper_bound,
            second_start_idx=upper_bound,
            confidence_level=confidence_level,
        )



class UnsatisfiableParameterizationRequestError(ValueError):
    """Raised when parameter with required properties cannot be found.

    May be due to inexistance of satisfactory parameter or due to violation of
    assumed monotonic relationship between parameter and focal property.
    """



from ..._impl import CalcMrcaUncertaintyRelUpperBoundWorstCase


# must inherit rather than assign due to failure of attribute lookup when
# pickling otherwise
class CalcMrcaUncertaintyRelUpperBound(
    CalcMrcaUncertaintyRelUpperBoundWorstCase
):
    pass



import random
import string

import pytest

from hstrat._auxiliary_lib import (
    parse_from_numeral_system,
    render_to_numeral_system,
)


@pytest.fixture
def alphabet():
    return string.digits + string.ascii_lowercase


def test_parse_from_numeral_system(alphabet):
    assert parse_from_numeral_system("0", alphabet) == 0
    assert parse_from_numeral_system("1", alphabet) == 1
    assert parse_from_numeral_system("a", alphabet) == 10
    assert parse_from_numeral_system("z", alphabet) == 35
    assert parse_from_numeral_system("10", alphabet) == 36
    assert parse_from_numeral_system("2s", alphabet) == 100
    assert parse_from_numeral_system("2n9c", alphabet) == 123456


def test_render_and_parse_numeral_system_consistency(alphabet):
    for __ in range(100):
        num = random.randint(0, 1000000)
        base_num = render_to_numeral_system(num, alphabet)
        assert parse_from_numeral_system(base_num, alphabet) == num

    for __ in range(100):
        test_digits = "".join([random.choice(alphabet) for j in range(10)])
        base_num = parse_from_numeral_system(test_digits, alphabet)
        assert render_to_numeral_system(
            base_num, alphabet
        ) == test_digits.lstrip("0")


def test_hex_parsing():
    assert parse_from_numeral_system("0", "0123456789abcdef") == int("0", 16)
    assert parse_from_numeral_system("1", "0123456789abcdef") == int("1", 16)
    assert parse_from_numeral_system("0a3", "0123456789abcdef") == int(
        "0a3", 16
    )
    assert parse_from_numeral_system("a3", "0123456789abcdef") == int("a3", 16)
    assert parse_from_numeral_system("ff", "0123456789abcdef") == int("ff", 16)


def test_octal_parsing():
    assert parse_from_numeral_system("0", "01234567") == int("0", 8)
    assert parse_from_numeral_system("1", "01234567") == int("1", 8)
    assert parse_from_numeral_system("052", "01234567") == int("052", 8)
    assert parse_from_numeral_system("52", "01234567") == int("52", 8)
    assert parse_from_numeral_system("777", "01234567") == int("777", 8)


def test_binary_parsing():
    assert parse_from_numeral_system("0", "01") == int("0", 2)
    assert parse_from_numeral_system("1", "01") == int("1", 2)
    assert parse_from_numeral_system("0101101", "01") == int("0101101", 2)
    assert parse_from_numeral_system("101101", "01") == int("101101", 2)
    assert parse_from_numeral_system("111", "01") == int("111", 2)


def test_decimal_parsing():
    assert parse_from_numeral_system("0", "0123456789") == int("0", 10)
    assert parse_from_numeral_system("1", "0123456789") == int("1", 10)
    assert parse_from_numeral_system("0123", "0123456789") == int("0123", 10)
    assert parse_from_numeral_system("123", "0123456789") == int("123", 10)
    assert parse_from_numeral_system("456", "0123456789") == int("456", 10)



import typing

from ...genome_instrumentation import HereditaryStratigraphicColumn


def iter_mutual_ranks(
    first: HereditaryStratigraphicColumn,
    second: HereditaryStratigraphicColumn,
    compare: bool = False,
    *,
    first_start_idx: int = 0,
    second_start_idx: int = 0,
) -> typing.Union[typing.Iterator[int], typing.Tuple[int, bool]]:
    """Iterate over ranks with matching strata between columns in ascending
    order."""
    # helper setup
    try:
        first_iter = first._stratum_ordered_store.IterRankDifferentiaZip(
            get_rank_at_column_index=first.GetRankAtColumnIndex,
            start_column_index=first_start_idx,
        )
        second_iter = second._stratum_ordered_store.IterRankDifferentiaZip(
            get_rank_at_column_index=second.GetRankAtColumnIndex,
            start_column_index=second_start_idx,
        )
    except AttributeError:
        first_iter = zip(
            first.IterRetainedRanks(), first.IterRetainedDifferentia()
        )
        second_iter = zip(
            second.IterRetainedRanks(), second.IterRetainedDifferentia()
        )
    first_cur_rank, first_cur_differentia = next(first_iter)
    second_cur_rank, second_cur_differentia = next(second_iter)

    # a.k.a.
    # while (
    #     first_column_idx < first.GetNumStrataRetained()
    #     and second_column_idx < second.GetNumStrataRetained()
    # ):
    try:
        while True:
            if first_cur_rank == second_cur_rank:
                # strata at same rank can be compared
                if first_cur_differentia == second_cur_differentia:
                    # matching differentiae at the same rank,
                    # store rank and keep searching for mismatch
                    yield (first_cur_rank, True) if compare else first_cur_rank
                    # advance first
                    first_cur_rank, first_cur_differentia = next(first_iter)
                    # advance second
                    second_cur_rank, second_cur_differentia = next(second_iter)
                else:
                    # mismatching differentiae at the same rank
                    yield (
                        first_cur_rank,
                        False,
                    ) if compare else first_cur_rank
            elif first_cur_rank < second_cur_rank:
                # current stratum on first column older than on second column
                # advance to next-newer stratum on first column
                first_cur_rank, first_cur_differentia = next(first_iter)
            elif first_cur_rank > second_cur_rank:
                # current stratum on second column older than on first column
                # advance to next-newer stratum on second column
                second_cur_rank, second_cur_differentia = next(second_iter)
    except StopIteration:
        return



import typing

import pandas as pd

from ._coerce_to_polars import coerce_to_polars
from ._render_polars_snapshot import render_polars_snapshot


def render_pandas_snapshot(
    df: pd.DataFrame,
    df_name: str = "",
    display: typing.Callable = lambda x: x,
) -> typing.Optional[str]:
    """Render an all-columns snapshot of the head of a Polars DataFrame.

    Parameters
    ----------
    df : pl.DataFrame
        The DataFrame to render.
    df_name : str, default ""
        The name of the DataFrame to render.
    display : Callable, optional
        The function to use to display the rendered DataFrame, e.g., print,
        logging.info, etc.

    Returns
    -------
    Optional[str]
        The rendered DataFrame, as a string, if `display` is not used.
    """
    return render_polars_snapshot(
        coerce_to_polars(df.head()), df_name=df_name, display=display
    )



import functools
import logging
import typing

import pandas as pd
import polars as pl

from ._coerce_to_pandas import (
    _supported_iterables,
    _supported_mappings,
    coerce_to_pandas,
)
from ._coerce_to_polars import coerce_to_polars
from ._warn_once import warn_once

DataFrame_T = typing.TypeVar("DataFrame_T", pd.DataFrame, pl.DataFrame)
Series_T = typing.TypeVar("Series_T", pd.Series, pl.Series)


def _detect_pandas(arg: typing.Any, recurse: bool) -> bool:
    """Implementation detail for delegate_polars_implementation.

    If `recurse` is True, then this function will recursively check for Polars
    members in mappings and iterables.
    """
    if isinstance(arg, (pd.DataFrame, pd.Series)):
        return True
    elif isinstance(arg, (pl.DataFrame, pl.Series, str)):
        return False
    elif recurse and isinstance(arg, _supported_mappings):
        return any(_detect_pandas(v, recurse) for v in arg.values())
    elif recurse and isinstance(arg, _supported_iterables):
        return any(_detect_pandas(ele, recurse) for ele in arg)
    elif isinstance(arg, (typing.Mapping, typing.Iterable)):
        warn_once(
            f"Arguments of type '{type(arg).__name__}' "
            "cannot be checked for Polars members.",
        )
    else:
        return False


def _detect_polars(arg: typing.Any, recurse: bool) -> bool:
    """Implementation detail for delegate_polars_implementation.

    If `recurse` is True, then this function will recursively check for Polars
    members in mappings and iterables.
    """
    if isinstance(arg, (pl.DataFrame, pl.Series)):
        return True
    elif isinstance(arg, (pd.DataFrame, pd.Series, str)):
        return False
    elif recurse and isinstance(arg, _supported_mappings):
        return any(_detect_polars(v, recurse) for v in arg.values())
    elif recurse and isinstance(arg, _supported_iterables):
        return any(_detect_polars(ele, recurse) for ele in arg)
    elif recurse and isinstance(arg, (typing.Mapping, typing.Iterable)):
        warn_once(
            f"Arguments of type '{type(arg).__name__}' "
            "cannot be checked for Pandas members.",
        )
    else:
        return False


def delegate_polars_implementation(
    polars_impl: typing.Optional[typing.Callable] = None,
    recurse_type_checks: bool = False,
) -> typing.Callable:
    """Decorates `pandas_impl` to either (1) dynamically dispatches calls using
    polars arguments to a supplied polars implementation or (2) coerces
    arguments to Pandas and calls `pandas_impl`.

    Calls without Polars arguments are dispatched to `pandas_impl` without any
    coercion.

    If `recurse_type_checks` is True, then coercion and Polars/Pandas
    detection happens recursively, checking iterables and/or mappings.

    Raises
    ------
    TypeError
        If mixing Pandas and Polars arguments is detected.
    """
    recurse = recurse_type_checks
    coerce_to_pandas_ = functools.partial(coerce_to_pandas, recurse=recurse)
    coerce_to_polars_ = functools.partial(coerce_to_polars, recurse=recurse)
    detect_pandas_ = functools.partial(_detect_pandas, recurse=recurse)
    detect_polars_ = functools.partial(_detect_polars, recurse=recurse)

    def decorator(pandas_impl: typing.Callable) -> typing.Callable:
        @functools.wraps(pandas_impl)
        def delegating_function(*args, **kwargs) -> typing.Any:

            any_pandas = any(map(detect_pandas_, (*args, *kwargs.values())))
            any_polars = any(map(detect_polars_, (*args, *kwargs.values())))
            logging.info("begin delgate_polars_implementation")
            logging.info("- detected {any_pandas=} {any_polars=}")

            if any_pandas and any_polars:
                raise TypeError("mixing pandas and polars types is disallowed")
            elif any_polars and polars_impl is not None:
                logging.info("- delegating to polars implementation")
                return polars_impl(*args, **kwargs)
            else:
                logging.info("- coercing args to pandas (possible no-op)")
                args = [*map(coerce_to_pandas_, args)]
                logging.info("- coercing kwargs to pandas (possible no-op)")
                kwargs = {
                    kw: coerce_to_pandas_(arg) for kw, arg in kwargs.items()
                }

                logging.info("- dispatching to pandas implementation")
                pandas_retval = pandas_impl(*args, **kwargs)

                if any_polars:
                    logging.info("- coercing return value to polars")
                    return coerce_to_polars_(pandas_retval)
                else:
                    return pandas_retval

        if delegating_function.__doc__ is not None:
            delegating_function.__doc__ += (
                "\n\nThis function also accepts a polars.DataFrame, for which "
                f"there is {'not ' if polars_impl is None else ''} a separate "
                "delegated implementation."
            )

        return delegating_function

    return decorator



import typing

from ._PerfectBacktrackHandle import PerfectBacktrackHandle


def iter_perfect_backtrack_lineage(
    handle: PerfectBacktrackHandle,
) -> typing.Iterator[PerfectBacktrackHandle]:
    """Iterate up `PerfectBacktrackHandle` breadcrumbs for line of descent
    ending in `handle`, including `handle`."""
    while True:
        yield handle

        if handle.parent is None:
            break
        else:
            handle = handle.parent



#!/usr/bin/env python3

from hstrat import hstrat

if __name__ == "__main__":

    for target_exact_absolute_mrca_uncertainty in (8, 127):
        for algo in (
            hstrat.geom_seq_nth_root_tapered_algo,
            hstrat.fixed_resolution_algo,
            hstrat.recency_proportional_resolution_algo,
        ):
            print("stratum retention algorithm", algo.__name__)
            print(f"{target_exact_absolute_mrca_uncertainty=}")

            stratum_retention_policy = algo.Policy(
                parameterizer=hstrat.PropertyAtMostParameterizer(
                    target_value=target_exact_absolute_mrca_uncertainty,
                    policy_evaluator=hstrat.MrcaUncertaintyAbsExactEvaluator(
                        at_num_strata_deposited=256,
                        at_rank=0,
                    ),
                    param_lower_bound=1,
                    param_upper_bound=1024,
                ),
            )
            print("   parameterized policy", stratum_retention_policy)

            provided_exact_absolute_mrca_uncertainty = (
                stratum_retention_policy.CalcMrcaUncertaintyAbsExact(
                    first_num_strata_deposited=256,
                    second_num_strata_deposited=256,
                    actual_rank_of_mrca=0,
                )
            )
            print(f"   {provided_exact_absolute_mrca_uncertainty=}")
            print()



import numpy as np

from hstrat._auxiliary_lib import is_subset


def test_empty_subset_and_superset():
    subset = np.array([])
    superset = np.array([])
    assert is_subset(subset, superset)


def test_empty_subset_not_empty_superset():
    subset = np.array([])
    superset = np.array([1, 2, 3])
    assert is_subset(subset, superset)


def test_not_empty_subset_empty_superset():
    subset = np.array([1, 2, 3])
    superset = np.array([])
    assert not is_subset(subset, superset)


def test_subset_not_a_subset_of_superset():
    subset = np.array([1, 2, 3, 4])
    superset = np.array([1, 2, 3])
    assert not is_subset(subset, superset)


def test_subset_is_a_subset_of_superset():
    subset = np.array([1, 2, 3])
    superset = np.array([1, 2, 3, 4])
    assert is_subset(subset, superset)



from hstrat._auxiliary_lib import get_package_name


def test_get_package_name():
    assert get_package_name("baz") == ""
    assert get_package_name("foo.bar") == "foo"
    assert get_package_name("foo.baz.bar") == "foo.baz"



import numpy as np
import pandas as pd


def get_nullable_vals(series: pd.Series) -> np.ndarray:
    """Get the integer underlying values in a Pandas Series with nullable
    integer dtype.

    Parameters
    ----------
    series : pd.Series
        The Pandas Series to get the null values for.

    Returns
    -------
    np.ndarray
        A 1-dimensional NumPy ndarray of the same dtype as the input series.
        The i-th element of the array is the underlying value at the i-th
        position in the input series.

    Notes
    -----
    This function returns a direct view into the Series data, so no copy is
    made. Changes to the returned array will propagate to the Series object's
    underlying values, and vice versa.
    """
    return series._data.array._data



import base64


# adapted from https://stackoverflow.com/a/45928164
def is_base64(data: str) -> bool:
    try:
        if isinstance(data, str):
            # If there's any unicode here, an exception will be thrown and the function will return false
            data_bytes = bytes(data, "ascii")
        elif isinstance(data, bytes):
            data_bytes = data
        else:
            raise ValueError("Argument must be string or bytes")
        return base64.b64encode(base64.b64decode(data_bytes)) == data_bytes
    except UnicodeEncodeError:
        return False



def bit_floor(n: int) -> int:
    """Calculate the largest power of two not greater than a
    non-negative integer n.

    If zero, returns zero.
    """
    assert n >= 0
    if n:
        # see https://stackoverflow.com/a/14267825/17332200
        # cast to int to make robust to numpy.int32, numpy.int64, etc.
        exp = int(n // 2).bit_length()
        res = 1 << exp
        return res
    else:
        return 0



import pandas as pd

from ._alifestd_count_leaf_nodes import alifestd_count_leaf_nodes


def alifestd_count_inner_nodes(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> int:
    """Count how many non-leaf nodes are contained in phylogeny."""

    num_leaves = alifestd_count_leaf_nodes(phylogeny_df)
    res = len(phylogeny_df) - num_leaves
    assert res >= 0
    return res



"""Implementation for `GarbageCollectingPhyloTracker`."""

from ._discern_referenced_rows import _discern_referenced_rows

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "_discern_referenced_rows",
]



import typing

import pandas as pd
import polars as pl

from ...._auxiliary_lib import (
    AnyTreeFastPreOrderIter,
    anytree_iterative_deepcopy,
)
from ...priors import ArbitraryPrior
from ...priors._detail import PriorBase
from .._impl import TrieInnerNode, TrieLeafNode
from ._detail import TriePostprocessorBase


def _call_anytree(
    ftor: "AssignOriginTimeSampleNaiveTriePostprocessor",
    trie: TrieInnerNode,
    progress_wrap: typing.Callable,
) -> TrieInnerNode:
    """Implementation detail."""
    for node in progress_wrap(AnyTreeFastPreOrderIter(trie)):
        if node.is_leaf:
            setattr(node, ftor._assigned_property, node.rank)
            assert isinstance(node, TrieLeafNode)
        elif node.parent is None:
            setattr(node, ftor._assigned_property, 0)
        else:
            interval_sample = ftor._prior.SampleIntervalConditionedValue(
                node.rank,
                min(
                    (
                        child.rank
                        for child in node.children
                        if not child.is_leaf
                    ),
                    default=node.rank + 1,
                ),  # endpoint is exclusive
            )
            setattr(
                node,
                ftor._assigned_property,
                min(
                    interval_sample,
                    min(
                        (child.rank for child in node.children),
                        default=interval_sample,
                    ),
                ),
            )

        assert hasattr(node, ftor._assigned_property)

    return trie


class AssignOriginTimeSampleNaiveTriePostprocessor(TriePostprocessorBase):
    """Functor to assign origin time property to trie nodes sampled between the
    node's rank and the minimum rank among its children.

    A prior may be provided to customize sampling distribution used.
    """

    _assigned_property: str  # property name for assigned origin time
    _prior: PriorBase  # prior expectation for ancestor origin times

    def __init__(
        self: "AssignOriginTimeSampleNaiveTriePostprocessor",
        prior: PriorBase = ArbitraryPrior(),  # ok as kwarg; immutable
        assigned_property: str = "origin_time",
    ) -> None:
        """Initialize functor instance.

        Parameters
        ----------
        prior : PriorBase, default ArbitraryPrior()
            Prior distribution of ancestor origin times.

            Used to calculate interval samples.
        assigned_property : str, default "origin_time"
            The property name for the assigned origin time.
        """
        self._assigned_property = assigned_property
        self._prior = prior

    def __call__(
        self: "AssignOriginTimeSampleNaiveTriePostprocessor",
        trie: TrieInnerNode,
        p_differentia_collision: float,
        mutate: bool = False,
        progress_wrap: typing.Callable = lambda x: x,
    ) -> TrieInnerNode:
        """Assign origin times to trie nodes.

        Parameters
        ----------
        trie : TrieInnerNode
            The input trie to be postprocessed.
        p_differentia_collision : float
            Probability of a randomly-generated differentia matching an
            existing differentia.

            Not used in the current implementation.
        mutate : bool, default False
            Are side effects on the input argument `trie` allowed?
        progress_wrap : typing.Callable, optional
            Pass tqdm or equivalent to report progress.

        Returns
        -------
        TrieInnerNode
            The postprocessed trie with assigned origin times.
        """
        if isinstance(trie, TrieInnerNode):
            if not mutate:
                trie = anytree_iterative_deepcopy(
                    trie, progress_wrap=progress_wrap
                )
            return _call_anytree(
                self,
                trie,
                progress_wrap=progress_wrap,
            )
        elif isinstance(trie, (pl.DataFrame, pd.DataFrame)):
            raise NotImplementedError  # pragma: no cover
        else:
            raise TypeError  # pragma: no cover



def count_trailing_zeros(n: int) -> int:
    """Count the number of trailing zeros in the binary representation of the
    absolute value of an integer.

    Trailing zeros are contiguous unset bits before the first set bit at the
    least significant (rightmost) position. For example, the number 8
    (1000 in binary) has 3 trailing zeros. Returns 0 trailing bits for 0.

    Examples
    --------
    >>> count_trailing_zeros(8)
    3

    >>> count_trailing_zeros(0)
    0
    """
    return (n & -n).bit_length() - 1 if n else 0



import typing

from .._detail import PolicySpecBase


class PolicySpec(PolicySpecBase):
    """Contains all policy parameters, if any."""

    _recency_proportional_resolution: int

    def __init__(
        self: "PolicySpec",
        recency_proportional_resolution: int,
    ) -> None:
        """Construct the policy spec.

        Parameters
        ----------
        recency_proportional_resolution : int, optional
            The desired minimum number of intervals between the MRCA and the
            deeper compared column to be able to be distinguished between. The
            uncertainty of MRCA rank estimates provided under the MRCA-recency-
            proportional resolution policy will scale as the actual
            phylogenetic depth of the MRCA divided by
            recency_proportional_resolution.
        """
        self._recency_proportional_resolution = recency_proportional_resolution

    def __eq__(self: "PolicySpec", other: typing.Any) -> bool:
        return isinstance(other, self.__class__) and (
            self._recency_proportional_resolution,
        ) == (other._recency_proportional_resolution,)

    def __repr__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoIdentifier()
        }.{
            PolicySpec.__qualname__
        }(recency_proportional_resolution={
            self._recency_proportional_resolution
        })"""

    def __str__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoTitle()
        } (resolution: {
            self._recency_proportional_resolution
        })"""

    def GetEvalCtor(self: "PolicySpec") -> str:
        return f"hstrat.{self!r}"

    def GetRecencyProportionalResolution(self: "PolicySpec") -> int:
        return self._recency_proportional_resolution

    @staticmethod
    def GetAlgoIdentifier() -> str:
        """Get programatic name for underlying retention algorithm."""
        return __package__.split(".")[-1]

    @staticmethod
    def GetAlgoTitle() -> str:
        """Get human-readable name for underlying retention algorithm."""
        return "Recency-proportional Resolution Stratum Retention Algorithm"



import typing

from ...._detail import PolicyCouplerBase
from ...._impl import GenDropRanksFromPredKeepRank
from ..._PolicySpec import PolicySpec
from ..._impl import calc_provided_uncertainty


class _PredKeepRank:
    """Functor to implement the depth-proportional resolution stratum retention
    policy, for use with HereditaryStratigraphicColumn.

    This functor enacts the depth-proportional resolution policy by specifying
    the set of strata ranks that should be purged from a hereditary
    stratigraphic column when the nth stratum is deposited.
    """

    def __init__(
        self: "_PredKeepRank",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "_PredKeepRank", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "_PredKeepRank",
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        stratum_rank: int,
    ) -> bool:
        """Decide if a stratum within the stratagraphic column should be
        retained or purged.

        Every time a new stratum is deposited, this method is called on each
        stratum present in a HereditaryStratigraphicColumn to determine whether
        it should be retained. Strata that return False are immediately purged
        from the column, meaning that for a stratum to persist it must earn a
        True result from this method each and every time a new stratum is
        deposited.

        Parameters
        ----------
        policy: Policy
            Policy this functor enacts.
        stratum_rank : int
            The number of strata that were deposited before the stratum under
            consideration for retention.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.

        Returns
        -------
        bool
            True if the stratum should be retained, False otherwise.

        See Also
        --------
        depth_proportional_resolution_algo:
            For details on the rationale, implementation, and guarantees of the
            depth-proportional resolution stratum retention policy.
        """
        spec = policy.GetSpec()
        guaranteed_resolution = spec.GetDepthProportionalResolution()

        # easy edge cases we must always retain
        if (
            # always retain newest stratum
            stratum_rank == num_stratum_depositions_completed
            # retain all strata until more than num_intervals are deposited
            or num_stratum_depositions_completed < guaranteed_resolution
        ):
            return True

        # +1 because of in-progress deposition
        provided_uncertainty = calc_provided_uncertainty(
            guaranteed_resolution,
            num_stratum_depositions_completed + 1,
        )
        return stratum_rank % provided_uncertainty == 0


FromPredKeepRank = GenDropRanksFromPredKeepRank(_PredKeepRank)



import os

import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_find_root_ids,
    alifestd_make_empty,
    alifestd_mark_oldest_root,
    alifestd_validate,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    original = phylogeny_df.copy()

    result = alifestd_mark_oldest_root(phylogeny_df)

    assert result["is_oldest_root"].sum() == np.sign(len(result))
    assert alifestd_validate(result)
    assert alifestd_find_root_ids(original) == alifestd_find_root_ids(result)
    assert (
        result.loc[result["is_oldest_root"], "id"]
        .isin(
            alifestd_find_root_ids(result),
        )
        .all()
    )


def test_empty():
    res = alifestd_mark_oldest_root(alifestd_make_empty())
    assert "is_oldest_root" in res.columns
    assert len(res) == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3],
            "ancestor_list": ["[None]", "[None]", "[0]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()

    result_df = alifestd_mark_oldest_root(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert not result_df.loc[1, "is_oldest_root"]
    assert result_df.loc[0, "is_oldest_root"]
    assert not result_df.loc[2, "is_oldest_root"]
    assert not result_df.loc[3, "is_oldest_root"]

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3],
            "origin_time": [0, 1, 2, 3],
            "ancestor_list": ["[None]", "[None]", "[0]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()

    result_df = alifestd_mark_oldest_root(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[1, "is_oldest_root"]
    assert not result_df.loc[0, "is_oldest_root"]
    assert not result_df.loc[2, "is_oldest_root"]
    assert not result_df.loc[3, "is_oldest_root"]

    if not mutate:
        assert original_df.equals(phylogeny_df)



import typing

import opytional as opyt

from .._impl import (
    CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
    CalcMrcaUncertaintyAbsUpperBoundPessimalRankBruteForce,
    CalcMrcaUncertaintyAbsUpperBoundWorstCase,
    CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
    CalcMrcaUncertaintyRelUpperBoundPessimalRankBruteForce,
    CalcMrcaUncertaintyRelUpperBoundWorstCase,
    CalcNumStrataRetainedUpperBoundWorstCase,
)
from ._PolicyCouplerBase import PolicyCouplerBase
from ._PolicySpecBase import PolicySpecBase
from ._UnsatisfiableParameterizationRequestError import (
    UnsatisfiableParameterizationRequestError,
)


class _CurryPolicy:
    """Helper class to enable the policy coupler to insert itself as the
    first argument to calls to implementation functors."""

    _policy: PolicyCouplerBase
    _ftor: typing.Callable

    def __init__(
        self: "_CurryPolicy",
        policy: PolicyCouplerBase,
        ftor: typing.Callable,
    ) -> None:
        self._policy = policy
        self._ftor = ftor

    def __eq__(self: "_CurryPolicy", other: typing.Any) -> bool:
        return isinstance(other, self.__class__) and (
            # don't compare policy to prevent infinite recursion
            self._ftor
            == other._ftor,
        )

    def __call__(self: "_CurryPolicy", *args, **kwargs) -> typing.Any:
        return self._ftor(self._policy, *args, **kwargs)


_ftor_type = typing.Type[typing.Callable]


def PolicyCouplerFactory(
    *,
    policy_spec_t: typing.Type,
    # enactment
    gen_drop_ranks_ftor_t: _ftor_type,
    # invariants
    calc_mrca_uncertainty_abs_upper_bound_ftor_t: _ftor_type = CalcMrcaUncertaintyAbsUpperBoundWorstCase,
    calc_mrca_uncertainty_abs_upper_bound_at_pessimal_rank_ftor_t: _ftor_type = CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
    calc_mrca_uncertainty_abs_upper_bound_pessimal_rank_ftor_t: _ftor_type = CalcMrcaUncertaintyAbsUpperBoundPessimalRankBruteForce,
    calc_mrca_uncertainty_rel_upper_bound_ftor_t: _ftor_type = CalcMrcaUncertaintyRelUpperBoundWorstCase,
    calc_mrca_uncertainty_rel_upper_bound_at_pessimal_rank_ftor_t: _ftor_type = CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
    calc_mrca_uncertainty_rel_upper_bound_pessimal_rank_ftor_t: _ftor_type = CalcMrcaUncertaintyRelUpperBoundPessimalRankBruteForce,
    calc_num_strata_retained_upper_bound_ftor_t: _ftor_type = CalcNumStrataRetainedUpperBoundWorstCase,
    # scrying
    calc_mrca_uncertainty_abs_exact_ftor_t: typing.Optional[_ftor_type] = None,
    calc_mrca_uncertainty_rel_exact_ftor_t: typing.Optional[_ftor_type] = None,
    calc_num_strata_retained_exact_ftor_t: typing.Optional[_ftor_type] = None,
    calc_rank_at_column_index_ftor_t: typing.Optional[_ftor_type] = None,
    iter_retained_ranks_ftor_t: typing.Optional[_ftor_type] = None,
) -> typing.Type[typing.Callable]:
    """Joins policy implementation functors into a single class that can be
    instantiated with particular policy specification parameters."""
    policy_spec_t_ = policy_spec_t

    class PolicyCoupler(
        PolicyCouplerBase,
    ):
        """Instantiate policy implementation for particular policy
        specification parameters."""

        _policy_spec: policy_spec_t_

        # invariants
        CalcMrcaUncertaintyAbsUpperBound: typing.Callable
        CalcMrcaUncertaintyAbsUpperBoundPessimalRank: typing.Callable
        CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank: typing.Callable
        CalcMrcaUncertaintyRelUpperBound: typing.Callable
        CalcMrcaUncertaintyRelUpperBoundAtPessimalRank: typing.Callable
        CalcMrcaUncertaintyRelUpperBoundPessimalRank: typing.Callable
        CalcNumStrataRetainedUpperBound: typing.Callable

        # scrying
        CalcMrcaUncertaintyAbsExact: typing.Optional[typing.Callable]
        CalcMrcaUncertaintyRelExact: typing.Optional[typing.Callable]
        CalcNumStrataRetainedExact: typing.Optional[typing.Callable]
        CalcRankAtColumnIndex: typing.Optional[typing.Callable]
        IterRetainedRanks: typing.Optional[typing.Callable]

        # enactment
        GenDropRanks: typing.Callable

        policy_spec_t: typing.Type = policy_spec_t_

        def __init__(
            self: "PolicyCoupler",
            *args,
            parameterizer: typing.Optional[
                typing.Callable[[typing.Type], typing.Optional[PolicySpecBase]]
            ] = None,
            policy_spec: typing.Optional[PolicySpecBase] = None,
            **kwargs,
        ):
            """Construct a PolicyCoupler instance.

            If policy_spec is not provided, all arguments are forwarded to
            policy spec initializer.
            """
            if policy_spec is not None:
                assert len(args) == len(kwargs) == 0
                assert parameterizer is None
                self._policy_spec = policy_spec
            elif parameterizer is not None:
                assert len(args) == len(kwargs) == 0
                assert policy_spec is None
                maybe_policy_spec = parameterizer(type(self))
                if maybe_policy_spec is None:
                    raise UnsatisfiableParameterizationRequestError(
                        repr(type(self)),
                        repr(parameterizer),
                    )
                else:
                    self._policy_spec = maybe_policy_spec
            else:
                assert parameterizer is None
                assert policy_spec is None
                self._policy_spec = policy_spec_t_(*args, **kwargs)

            # enactment
            self.GenDropRanks = _CurryPolicy(
                self,
                gen_drop_ranks_ftor_t(self._policy_spec),
            )

            # invariants
            self.CalcMrcaUncertaintyAbsUpperBound = _CurryPolicy(
                self,
                calc_mrca_uncertainty_abs_upper_bound_ftor_t(
                    self._policy_spec
                ),
            )
            self.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank = _CurryPolicy(
                self,
                calc_mrca_uncertainty_abs_upper_bound_at_pessimal_rank_ftor_t(
                    self._policy_spec,
                ),
            )
            self.CalcMrcaUncertaintyAbsUpperBoundPessimalRank = _CurryPolicy(
                self,
                calc_mrca_uncertainty_abs_upper_bound_pessimal_rank_ftor_t(
                    self._policy_spec,
                ),
            )
            self.CalcMrcaUncertaintyRelUpperBound = _CurryPolicy(
                self,
                calc_mrca_uncertainty_rel_upper_bound_ftor_t(
                    self._policy_spec
                ),
            )
            self.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank = _CurryPolicy(
                self,
                calc_mrca_uncertainty_rel_upper_bound_at_pessimal_rank_ftor_t(
                    self._policy_spec
                ),
            )
            self.CalcMrcaUncertaintyRelUpperBoundPessimalRank = _CurryPolicy(
                self,
                calc_mrca_uncertainty_rel_upper_bound_pessimal_rank_ftor_t(
                    self._policy_spec
                ),
            )
            self.CalcNumStrataRetainedUpperBound = _CurryPolicy(
                self,
                calc_num_strata_retained_upper_bound_ftor_t(self._policy_spec),
            )

            # scrying
            self.CalcMrcaUncertaintyAbsExact = opyt.apply_if(
                calc_mrca_uncertainty_abs_exact_ftor_t,
                lambda x: _CurryPolicy(self, x(self._policy_spec)),
            )
            self.CalcMrcaUncertaintyRelExact = opyt.apply_if(
                calc_mrca_uncertainty_rel_exact_ftor_t,
                lambda x: _CurryPolicy(self, x(self._policy_spec)),
            )
            self.CalcNumStrataRetainedExact = opyt.apply_if(
                calc_num_strata_retained_exact_ftor_t,
                lambda x: _CurryPolicy(self, x(self._policy_spec)),
            )
            self.CalcRankAtColumnIndex = opyt.apply_if(
                calc_rank_at_column_index_ftor_t,
                lambda x: _CurryPolicy(self, x(self._policy_spec)),
            )
            self.IterRetainedRanks = opyt.apply_if(
                iter_retained_ranks_ftor_t,
                lambda x: _CurryPolicy(self, x(self._policy_spec)),
            )

        def __eq__(
            self: "PolicyCoupler",
            other: typing.Any,
        ) -> bool:
            if issubclass(
                other.__class__,
                PolicyCouplerBase,
            ):
                return (
                    self._policy_spec,
                    self.GenDropRanks,
                    self.CalcMrcaUncertaintyAbsUpperBound,
                    self.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
                    self.CalcMrcaUncertaintyAbsUpperBoundPessimalRank,
                    self.CalcMrcaUncertaintyRelUpperBound,
                    self.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
                    self.CalcMrcaUncertaintyRelUpperBoundPessimalRank,
                    self.CalcNumStrataRetainedUpperBound,
                    self.CalcMrcaUncertaintyAbsExact,
                    self.CalcMrcaUncertaintyRelExact,
                    self.CalcNumStrataRetainedExact,
                    self.CalcRankAtColumnIndex,
                    self.IterRetainedRanks,
                ) == (
                    other._policy_spec,
                    other.GenDropRanks,
                    other.CalcMrcaUncertaintyAbsUpperBound,
                    other.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
                    other.CalcMrcaUncertaintyAbsUpperBoundPessimalRank,
                    other.CalcMrcaUncertaintyRelUpperBound,
                    other.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
                    other.CalcMrcaUncertaintyRelUpperBoundPessimalRank,
                    other.CalcNumStrataRetainedUpperBound,
                    other.CalcMrcaUncertaintyAbsExact,
                    other.CalcMrcaUncertaintyRelExact,
                    other.CalcNumStrataRetainedExact,
                    other.CalcRankAtColumnIndex,
                    other.IterRetainedRanks,
                )
            else:
                return False

        def __hash__(self: "PolicyCoupler") -> int:
            """Hash object instance."""
            return hash(self._policy_spec)

        def __repr__(self: "PolicyCoupler") -> str:
            return f"""{
                self._policy_spec.GetAlgoIdentifier()
            }.{
                PolicyCoupler.__qualname__
            }(policy_spec={
                self._policy_spec
            !r})"""

        def __str__(self: "PolicyCoupler") -> str:
            return str(self._policy_spec)

        def GetEvalCtor(self: "PolicyCoupler") -> str:
            return f"""hstrat.{
                self._policy_spec.GetAlgoIdentifier()
            }.Policy(policy_spec={
                self._policy_spec.GetEvalCtor()
            })"""

        def GetSpec(self: "PolicyCoupler") -> policy_spec_t_:
            """Get policy's parameter specification."""
            return self._policy_spec

        def WithoutCalcRankAtColumnIndex(
            self: "PolicyCoupler",
        ) -> "PolicyCoupler":
            """Make a copy of this policy instance with CalcRankAtColumnIndex
            disabled.

            Useful to prevent optimization-out of strata rank number storage in
            a stratum ordered store backing a hereditary stratigraphic column.
            """
            type_ = PolicyCouplerFactory(
                policy_spec_t=policy_spec_t_,
                # enactment
                gen_drop_ranks_ftor_t=gen_drop_ranks_ftor_t,
                # invariants
                calc_mrca_uncertainty_abs_upper_bound_ftor_t=calc_mrca_uncertainty_abs_upper_bound_ftor_t,
                calc_mrca_uncertainty_abs_upper_bound_at_pessimal_rank_ftor_t=calc_mrca_uncertainty_abs_upper_bound_at_pessimal_rank_ftor_t,
                calc_mrca_uncertainty_abs_upper_bound_pessimal_rank_ftor_t=calc_mrca_uncertainty_abs_upper_bound_pessimal_rank_ftor_t,
                calc_mrca_uncertainty_rel_upper_bound_ftor_t=calc_mrca_uncertainty_rel_upper_bound_ftor_t,
                calc_mrca_uncertainty_rel_upper_bound_at_pessimal_rank_ftor_t=calc_mrca_uncertainty_rel_upper_bound_at_pessimal_rank_ftor_t,
                calc_mrca_uncertainty_rel_upper_bound_pessimal_rank_ftor_t=calc_mrca_uncertainty_rel_upper_bound_pessimal_rank_ftor_t,
                calc_num_strata_retained_upper_bound_ftor_t=calc_num_strata_retained_upper_bound_ftor_t,
                # scrying
                calc_mrca_uncertainty_abs_exact_ftor_t=calc_mrca_uncertainty_abs_exact_ftor_t,
                calc_mrca_uncertainty_rel_exact_ftor_t=calc_mrca_uncertainty_rel_exact_ftor_t,
                calc_num_strata_retained_exact_ftor_t=calc_num_strata_retained_exact_ftor_t,
                calc_rank_at_column_index_ftor_t=None,
                iter_retained_ranks_ftor_t=iter_retained_ranks_ftor_t,
            )

            # propagate any glossing over over implementation details
            type_.__module__ = PolicyCoupler.__module__
            type_.__name__ = PolicyCoupler.__name__
            type_.__qualname__ = PolicyCoupler.__qualname__

            return type_(
                policy_spec=self._policy_spec,
            )

    return PolicyCoupler



import os
import subprocess

assets = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets")


def test_surface_postprocess_trie_cli_version():
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_postprocess_trie",
            "--version",
        ],
        check=True,
    )


def test_surface_postprocess_trie_cli_csv():
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_unpack_reconstruct",
            "/tmp/hstrat_unpack_surface_reconstruct_.csv",
        ],
        check=True,
        input=f"{assets}/packed.csv".encode(),
    )
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_postprocess_trie",
            "/tmp/hstrat_surface_postprocess_trie.csv",
        ],
        check=True,
        input="/tmp/hstrat_unpack_surface_reconstruct_.csv".encode(),
    )


def test_surface_postprocess_trie_cli_parquet():
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_unpack_reconstruct",
            "/tmp/hstrat_unpack_surface_reconstruct_.pqt",
            "--shrink-dtypes",
            "--exploded-slice-size",
            "50_000_000",
        ],
        check=True,
        input=f"{assets}/packed.csv".encode(),
    )
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_postprocess_trie",
            "/tmp/hstrat_surface_postprocess_trie.pqt",
            "--shrink-dtypes",
            "--trie-postprocessor",
            "hstrat.NopTriePostprocessor()",
        ],
        check=True,
        input="/tmp/hstrat_unpack_surface_reconstruct_.pqt".encode(),
    )



import pytest

from hstrat._auxiliary_lib import warn_once


@pytest.mark.filterwarnings("error")
def test_warn_once():
    msg1 = "asdf"
    msg2 = "hello world"

    # should warn
    with pytest.warns(UserWarning, match=msg1):
        warn_once(msg1)

    # repeat shouldn't warn
    warn_once(msg1)

    # new message should warn
    with pytest.warns(UserWarning, match=msg2):
        warn_once(msg2)



from ..._detail import PolicyCouplerBase
from ...geom_seq_nth_root_algo import Policy as gsnra_Policy
from ...recency_proportional_resolution_algo import Policy as rpra_Policy
from ._calc_provided_degree import calc_provided_degree
from ._calc_provided_resolution import calc_provided_resolution


def pick_policy(
    size_curb: int,
    num_stratum_depositions_completed: int,
) -> PolicyCouplerBase:
    """Helper that dispatches current retention policy after n depositions."""
    resolution = calc_provided_resolution(
        size_curb,
        num_stratum_depositions_completed,
    )

    if resolution >= 0:
        return rpra_Policy(
            resolution,
        )
    else:
        interspersal = 2
        degree = calc_provided_degree(
            size_curb,
            interspersal,
            num_stratum_depositions_completed,
        )
        return gsnra_Policy(
            degree,
            interspersal,
        )



import pandas as pd

from ._capitalize_n import capitalize_n


def as_nullable_type(series: pd.Series) -> pd.Series:
    """Convert the input Series to ensure a nullable dtype.

    If the Series' type is already nullable, it will be returned unchanged.

    Parameters
    ----------
    series : pandas.Series
        Input Series to be converted to a nullable data type.

    Returns
    -------
    pandas.Series
        A new Pandas Series object with a nullable dtype, converted from the input Series.

    Notes
    -----
    See <https://pandas.pydata.org/pandas-docs/version/1.5/user_guide/integer_na.html>
    and <https://pandas.pydata.org/pandas-docs/version/1.5/user_guide/boolean.html> for information on nullable integer and boolean types.

    Examples
    --------
    >>> import pandas as pd
    >>> from typing import Any
    >>> from ._capitalize_n import capitalize_n

    >>> data = {'A': [1, 2, 3], 'B': [True, False, True]}
    >>> df = pd.DataFrame(data)
    >>> nullable_df = df.apply(as_nullable_type)
    >>> nullable_df.dtypes
    A          Int64
    B    BooleanDtype
    dtype: object
    """
    if pd.api.types.is_bool_dtype(series.dtype):
        return series.astype("boolean")
    elif pd.api.types.is_unsigned_integer_dtype(series.dtype):
        return series.astype(capitalize_n(series.dtype.name, 2))
    elif pd.api.types.is_signed_integer_dtype(series.dtype):
        return series.astype(series.dtype.name.capitalize())
    else:
        return series



import typing

from ._iter_rank_cutoffs import iter_rank_cutoffs
from ._iter_rank_seps import iter_rank_seps
from ._iter_target_ranks import iter_target_ranks


def iter_rank_backstops(
    degree: int,
    interspersal: int,
    num_strata_deposited: int,
) -> typing.Iterator[int]:
    """Yield most ancient retained rank for each exponentially-spaced
    coverage target `pow` in ascending order.

    All subsequent ranks spaced forward by `_calc_rank_sep`
    positions through recency zero will be retained. Will monotonically
    increase with `num_strata_deposited` and be an even multiple of
    `_calc_rank_sep`.
    """
    for rank_cutoff, retained_ranks_sep, target_rank in zip(
        iter_rank_cutoffs(degree, interspersal, num_strata_deposited),
        iter_rank_seps(degree, interspersal, num_strata_deposited),
        iter_target_ranks(degree, num_strata_deposited),
    ):

        # round UP from rank_cutoff
        # to align evenly with retained_ranks_sep
        # adapted from https://stackoverflow.com/a/14092788
        min_retained_rank = rank_cutoff - (rank_cutoff % -retained_ranks_sep)
        assert min_retained_rank % retained_ranks_sep == 0
        assert min_retained_rank >= rank_cutoff

        # check that even with rounding up, we are still covering the
        # target rank
        # i.e., that the most ancient retained rank (the backstop rank)
        # falls before the target rank so that the target rank is
        # guaranteed within an _iter_rank_sep window
        assert min_retained_rank <= target_rank

        # more sanity checks on range of return value
        if num_strata_deposited == 0:
            assert min_retained_rank == 0
        else:
            assert 0 <= min_retained_rank <= num_strata_deposited - 1

        # backstop rank synonomous w/ the most ancient (min) retained rank
        yield min_retained_rank



from ...._auxiliary_lib import get_package_name, load_cppimportable_module

_impl_mod = load_cppimportable_module(
    "_build_tree_searchtable_cpp_impl", get_package_name(__name__)
)
placeholder_value = _impl_mod.placeholder_value
Records = _impl_mod.Records
collapse_dropped_unifurcations = _impl_mod.collapse_dropped_unifurcations
extend_tree_searchtable_cpp_from_exploded = (
    _impl_mod.extend_tree_searchtable_cpp_from_exploded
)
build_tree_searchtable_cpp_from_exploded = (
    _impl_mod.build_tree_searchtable_cpp_from_exploded
)
build_tree_searchtable_cpp_from_nested = (
    _impl_mod.build_tree_searchtable_cpp_from_nested
)
records_to_dict = _impl_mod.records_to_dict



import warnings

import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._alifestd_unfurl_lineage_asexual import alifestd_unfurl_lineage_asexual
from ._pairwise import pairwise


def alifestd_reroot_at_id_asexual(
    phylogeny_df: pd.DataFrame,
    new_root_id: int,
    mutate: bool = False,
) -> pd.DataFrame:
    """Reroot phylogeny, preserving topology.

    Reverses the descendant-to-ancestor relationships of all ancestors of the
    new root. Does not update branch_lengths or edge_lengths columns if
    present.

    Parameters
    ----------
    phylogeny_df : pandas.DataFrame
        The phylogeny as a dataframe in alife standard format.

        Must represent an asexual phylogeny.
    new_root_id : int
        The ID of the node to use as the new root of the phylogeny.
    mutate : bool, default False
        Are side effects on the input argument `phylogeny_df` allowed?

    Returns
    -------
    pandas.DataFrame
        The rerooted phylogeny in alife standard format.
    """

    if "branch_length" in phylogeny_df or "edge_length" in phylogeny_df:
        warnings.warn(
            "alifestd_reroot_at_id_asexual does not update branch length "
            "columns. Use `origin_time` to recalculate branch lengths for "
            "rerooted phylogeny."
        )

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)
    unfurled_lineage = alifestd_unfurl_lineage_asexual(
        phylogeny_df, new_root_id
    )

    # contiguous id implementation
    if alifestd_has_contiguous_ids(phylogeny_df):
        copy_to_slice = unfurled_lineage[1:]
        copy_from_slice = unfurled_lineage[:-1]
        phylogeny_df["ancestor_id"].to_numpy()[copy_to_slice] = phylogeny_df[
            "id"
        ].to_numpy()[copy_from_slice]

        phylogeny_df["ancestor_id"].to_numpy()[new_root_id] = phylogeny_df[
            "id"
        ].to_numpy()[new_root_id]

    # noncontiguous id implementation
    else:
        iloc_lookup = dict(
            zip(phylogeny_df["id"], np.arange(len(phylogeny_df)))
        )
        for ancestor_id, descendant_id in pairwise(reversed(unfurled_lineage)):
            iloc = iloc_lookup[ancestor_id]
            phylogeny_df["ancestor_id"].to_numpy()[iloc] = descendant_id

        new_root_iloc = iloc_lookup[new_root_id]
        phylogeny_df["ancestor_id"].to_numpy()[new_root_iloc] = phylogeny_df[
            "id"
        ].to_numpy()[new_root_iloc]

    # update ancestor list
    phylogeny_df["ancestor_list"] = alifestd_make_ancestor_list_col(
        phylogeny_df["id"],
        phylogeny_df["ancestor_id"],
    )
    return phylogeny_df



import itertools as it
import typing

import opytional as opyt
import pandas as pd

from ..._auxiliary_lib import demark


def compile_phylogeny_from_lineage_iters(
    population: typing.Iterable[typing.Iterable[typing.Any]],
) -> pd.DataFrame:
    """Compile phylogenetic history tracked using generic backtracking breadcrumbs.

    Parameters
    ----------
    population : iterable of iterable
        Ascending lineage iterators associated with extant population members.

    Returns
    -------
    pd.DataFrame
        Phylogenetic record in alife data standards format.
    """

    seen_handle_ids = set()

    records = list()
    for lineage_it1, lineage_it2 in map(it.tee, population):
        for descendant_handle, ancestor_handle in it.zip_longest(
            lineage_it1,
            it.islice(lineage_it2, 1, None),
        ):
            assert descendant_handle is not None
            if demark(descendant_handle) not in seen_handle_ids:
                seen_handle_ids.add(demark(descendant_handle))
                records.append(
                    {
                        **{
                            "id": demark(descendant_handle),
                            "ancestor_list": str(
                                [opyt.apply_if(ancestor_handle, demark)]
                            ),
                        },
                        **(
                            {}
                            if not hasattr(descendant_handle, "data")
                            else descendant_handle.data
                            if isinstance(descendant_handle.data, dict)
                            else {"data": descendant_handle.data}
                            if descendant_handle.data is not None
                            else {}
                        ),
                    }
                )

    return pd.DataFrame.from_dict(records)



import logging
import subprocess
import typing


# adapted from https://stackoverflow.com/a/47411778/17332200
def log_memory_usage(logger: typing.Callable = logging.info) -> None:
    """Log memory use."""
    try:
        message = (
            "memory usage:\n"
            + subprocess.check_output(
                [
                    "free",
                    "--human",
                    "--total",
                ],
                shell=False,
            ).decode()
        )
    except subprocess.CalledProcessError as e:
        message = f"logging memory use failed: {e}"

    logger(message)



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_make_empty,
    alifestd_sum_origin_time_deltas_asexual,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_empty():
    mt = alifestd_make_empty()
    mt["origin_time"] = None
    assert 0 == alifestd_sum_origin_time_deltas_asexual(mt)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "origin_time": [0, 10, 30],
        }
    )
    original_df = phylogeny_df.copy()
    assert 30 == alifestd_sum_origin_time_deltas_asexual(
        phylogeny_df,
        mutate=mutate,
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3],
            "ancestor_list": ["[0]", "[None]", "[0]", "[1]"],
            "origin_time": [20, 10, 30, 45],
        }
    )
    original_df = phylogeny_df.copy()
    assert 55 == alifestd_sum_origin_time_deltas_asexual(
        phylogeny_df,
        mutate=mutate,
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple3(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3],
            "ancestor_list": ["[None]", "[None]", "[0]", "[1]"],
            "origin_time": [20, 10, 30, 45],
        }
    )
    original_df = phylogeny_df.copy()
    assert 45 == alifestd_sum_origin_time_deltas_asexual(
        phylogeny_df,
        mutate=mutate,
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)



import typing
import warnings

import numpy as np
import pandas as pd

from ._alifestd_is_asexual import alifestd_is_asexual
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def _alifestd_coarsen_mask_asexual(
    phylogeny_df: pd.DataFrame,
    mask: pd.Series,  # boolean mask
    progress_wrap: typing.Callable,
) -> pd.DataFrame:

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    root_token_str = phylogeny_df["ancestor_list"].iat[0]
    assert root_token_str[0] == "[" and root_token_str[-1] == "]"
    assert not any(c.isdigit() for c in root_token_str)

    phylogeny_df.set_index("id", drop=False, inplace=True)

    coarsened_ids = {*phylogeny_df.loc[mask, "id"]}
    new_ancestor_ids = {}  # int id --> int ancestor_id
    for idx in progress_wrap(phylogeny_df.index):
        ancestor_id = phylogeny_df.at[idx, "ancestor_id"]

        if ancestor_id in coarsened_ids:
            new_ancestor_ids[idx] = ancestor_id
        elif ancestor_id in new_ancestor_ids:
            assert ancestor_id != idx
            ancestor_id = new_ancestor_ids[ancestor_id]
            new_ancestor_ids[idx] = ancestor_id

    res = phylogeny_df[mask].reset_index(drop=True)
    res["ancestor_id"] = (
        # non-exhaustive mapping,
        # see https://stackoverflow.com/a/41678874/17332200
        res["id"]
        .map(new_ancestor_ids)
        .fillna(res["id"])
        .astype(int)
    )

    res["ancestor_list"] = alifestd_make_ancestor_list_col(
        res["id"], res["ancestor_id"], root_token_str[1:-1]
    )

    return res


def _alifestd_coarsen_mask_sexual(
    phylogeny_df: pd.DataFrame,
    mask: pd.Series,  # boolean mask
    progress_wrap: typing.Callable,
) -> pd.DataFrame:

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    root_token_str = phylogeny_df["ancestor_list"].iat[0]
    assert not any(c.isdigit() for c in root_token_str)

    phylogeny_df.set_index("id", drop=False, inplace=True)

    coarsened_ids = {*phylogeny_df.loc[mask, "id"]}
    new_ancestor_lists = {}  # int id --> list[int] ancestor_ids
    for idx in progress_wrap(phylogeny_df.index):
        ancestor_list = phylogeny_df.at[idx, "ancestor_list"]
        ancestor_ids = alifestd_parse_ancestor_ids(ancestor_list)

        new_ancestor_ids = []
        for ancestor_id in ancestor_ids:
            if ancestor_id in coarsened_ids:
                new_ancestor_ids.append(ancestor_id)
            else:
                new_ancestor_ids.extend(new_ancestor_lists[ancestor_id])

        new_ancestor_lists[idx] = new_ancestor_ids

    res = phylogeny_df[mask].reset_index(drop=True)
    res["ancestor_list"] = (
        res["id"]
        .map(new_ancestor_lists)
        .apply(np.unique)  # sort and prevent duplicate values
        .apply(np.ndarray.tolist)
        .apply(str)
        .replace("[]", root_token_str)
    )

    # don't need update because not in sexual phylos
    assert "ancestor_id" not in phylogeny_df

    return res


def alifestd_coarsen_mask(
    phylogeny_df: pd.DataFrame,
    mask: pd.Series,  # boolean mask
    mutate: bool = False,
    progress_wrap: typing.Callable = lambda x: x,
) -> pd.DataFrame:
    """Pare record to bypass organisms outside mask.

    The root ancestor token will be adopted from phylogeny_df.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if "branch_length" in phylogeny_df or "edge_length" in phylogeny_df:
        warnings.warn(
            "alifestd_coarsen_mask does not update branch length columns. "
            "Use `origin_time` to recalculate branch lengths for coarsened "
            "phylogeny."
        )

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if len(phylogeny_df) <= 1:
        return phylogeny_df

    if alifestd_is_asexual(phylogeny_df):
        return _alifestd_coarsen_mask_asexual(
            phylogeny_df, mask, progress_wrap
        )
    else:
        return _alifestd_coarsen_mask_sexual(phylogeny_df, mask, progress_wrap)



import anytree
import pytest

from hstrat._auxiliary_lib import anytree_has_grandparent


@pytest.fixture
def tree():
    root = anytree.Node("A")
    b = anytree.Node("B", parent=root)
    c = anytree.Node("C", parent=b)
    anytree.Node("D", parent=c)
    e = anytree.Node("E", parent=root)
    anytree.Node("F", parent=e)
    anytree.Node("X", parent=root)
    return root


def test_node_has_grandparent(tree):
    assert anytree_has_grandparent(tree.children[0].children[0])
    assert anytree_has_grandparent(tree.children[1].children[0])


def test_root_node_has_no_grandparent(tree):
    assert not anytree_has_grandparent(tree)
    assert not anytree_has_grandparent(anytree.Node("root"))


def test_node_has_no_grandparent(tree):
    for child in tree.children:
        assert not anytree_has_grandparent(child)



import numbers
import typing

from ._intersect_ranges import intersect_ranges


def make_intersecting_subsets(
    ranges: typing.List[typing.Tuple[numbers.Real, numbers.Real]],
) -> typing.List[typing.List[typing.Tuple[numbers.Real, numbers.Real]]]:
    """Arrange ranges into groups such that all ranges within a group share at
    least one overlapping value.

    Composes groups such that intersected values are as low as possible. Put
    another way, ranges are preferentially associated with earlier groups.

    Parameters
    ----------
    ranges : list of tuples
        A list of tuples, each containing two real numbers representing the
        lwoer bound (inclusive) and upper bound (exclusive) of a range.

    Returns
    -------
    list of lists of tuples
        A list of lists, each containing tuples representing the intersecting
        ranges within a group.


    Examples
    --------
    >>> make_intersecting_subsets([(1, 3), (2, 4), (5, 6)])
    [[(1, 3), (2, 4)], [(5, 6)]]
    """

    subsets = []
    for range_ in sorted(ranges):
        for subset in subsets:
            if intersect_ranges([range_, *subset]):
                subset.append(range_)
                break
        else:
            subsets.append([range_])

    return subsets



from ._estimate_rank_of_mrca_maximum_likelihood import (
    estimate_rank_of_mrca_maximum_likelihood,
)
from ._estimate_rank_of_mrca_naive import estimate_rank_of_mrca_naive
from ._estimate_rank_of_mrca_unbiased import estimate_rank_of_mrca_unbiased

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "estimate_rank_of_mrca_maximum_likelihood",
    "estimate_rank_of_mrca_naive",
    "estimate_rank_of_mrca_unbiased",
]



import typing

import interval_search as inch

from ._jit import jit


def curried_binary_search_jit(predicate: typing.Callable[[int], bool]):
    """Find the positive integer threshold below which a search criteria is
    never satisfied and above which it is always satisfied.

    Currying allows for recursive calls that don't pass the predicate callable
    as an argument, which is necessary for compatibility with Numba jit
    compilation.

    Parameters
    ----------
    predicate : callable object
        Returns whether an integer value satisfies the search criteria.
    lower_bound : int
        Lower bound for the binary search, inclusive.
    upper_bound : int
        Upper bound for the binary search, inclusive.

    Returns
    -------
    searcher : Callable[[int, int], Optional[int]]
        The curried search function, which takes an optional integer
        `lower_bound` and `upper_bound` and returns the first integer value
        that satisfies the search criteria.

        If no value satisfies the search criteria, the curried search function will return None.
    """
    return inch.curried_binary_search(
        predicate, decorate_with=jit(nopython=True)
    )



import os

import alifedata_phyloinformatics_convert as apc
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_as_newick_asexual,
    alifestd_make_empty,
    alifestd_try_add_ancestor_id_col,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df)
    original = phylogeny_df.copy()

    result = alifestd_as_newick_asexual(phylogeny_df, label_key="id")
    assert original.equals(phylogeny_df)

    rosetta_tree = apc.RosettaTree.from_newick(result)
    reconstructed = rosetta_tree.as_alife
    reconstructed["taxon_label"].fillna(reconstructed["label"], inplace=True)
    reconstructed["taxon_label"] = reconstructed["taxon_label"].astype(int)
    reconstructed = alifestd_try_add_ancestor_id_col(reconstructed)

    assert len(reconstructed) == len(original)

    taxon_labels = dict(zip(reconstructed["id"], reconstructed["taxon_label"]))
    assert set(zip(original["id"], original["ancestor_id"])) == set(
        zip(
            reconstructed["taxon_label"],
            reconstructed["ancestor_id"].map(taxon_labels),
        )
    ), (original, result, rosetta_tree.as_newick)


def test_empty():
    res = alifestd_as_newick_asexual(alifestd_make_empty())
    assert res == ";"


@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "origin_time_delta": [3.1, 4.0, 1.0],
        },
    )
    original_df = phylogeny_df.copy()
    result = alifestd_as_newick_asexual(
        phylogeny_df,
        label_key=None,
        mutate=mutate,
    )
    assert result == "((:1):4):3.1;"

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 0, 1, 2],
            "ancestor_list": ["[0]", "[1]", "[None]", "[1]"],
        },
    )
    original_df = phylogeny_df.copy()
    result = alifestd_as_newick_asexual(
        phylogeny_df,
        label_key=None,
        mutate=mutate,
    )
    assert result == "(,());"

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple3(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [4, 0, 2, 3],
            "ancestor_list": ["[None]", "[None]", "[0]", "[4]"],
            "label": ["A", "B", "C", "D"],
        },
    )
    original_df = phylogeny_df.copy()
    result = alifestd_as_newick_asexual(
        phylogeny_df,
        label_key="label",
        mutate=mutate,
    )
    assert result == "(D)A;\n(C)B;"

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple4(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[None]", "[0]", "[0]", "[1]", "[0]"],
            "origin_time": [0, 1, 2, 5, 90],
        },
    )
    original_df = phylogeny_df.copy()
    result = alifestd_as_newick_asexual(
        phylogeny_df,
        mutate=mutate,
        label_key="id",
    )

    assert result == "(4:90,2:2,(3:4)1:1)0:0;"

    if not mutate:
        assert original_df.equals(phylogeny_df)



import typing

from interval_search import binary_search

from ...genome_instrumentation import HereditaryStratigraphicColumn
from .._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)
from ._calc_rank_of_last_retained_commonality_between_generic import (
    calc_rank_of_last_retained_commonality_between_generic,
)


def calc_rank_of_last_retained_commonality_between_bsearch(
    first: HereditaryStratigraphicColumn,
    second: HereditaryStratigraphicColumn,
    confidence_level: float,
) -> typing.Optional[int]:
    """Find rank of strata commonality before first strata disparity.

    Implementation detail. Provides optimized implementation for
    special case where both first and second use the perfect resolution
    stratum retention policy.
    """
    # both must have (effectively) used the perfect resolution policy
    assert not first.HasDiscardedStrata() and not second.HasDiscardedStrata()

    lower_bound = 0
    upper_bound = min(
        [
            first.GetNumStrataDeposited() - 1,
            second.GetNumStrataDeposited() - 1,
        ]
    )
    assert lower_bound <= upper_bound

    def differentia_at(
        which: HereditaryStratigraphicColumn,
        idx: int,
    ) -> int:
        return which.GetStratumAtColumnIndex(idx).GetDifferentia()

    def predicate(idx: int) -> bool:
        return differentia_at(first, idx) != differentia_at(second, idx)

    first_disparite_idx = binary_search(
        predicate,
        lower_bound,
        upper_bound,
    )

    collision_implausibility_threshold = calc_min_implausible_spurious_consecutive_differentia_collisions_between(
        first,
        second,
        significance_level=1.0 - confidence_level,
    )
    assert collision_implausibility_threshold > 0
    if first_disparite_idx is None:
        # no disparate strata found
        # fall back to calc_rank_of_last_retained_commonality_between_generic
        # to handle proper bookkeeping in this case while skipping most of
        # the search
        return calc_rank_of_last_retained_commonality_between_generic(
            first,
            second,
            first_start_idx=upper_bound,
            second_start_idx=upper_bound,
            confidence_level=confidence_level,
        )
    elif first_disparite_idx >= collision_implausibility_threshold:
        # disparate strata found, following some common strata
        # ...discount collision_implausibility_threshold - 1 common strata
        # as potential spurious differentia collisions
        # ... must also subtract 1 (canceling out -1 above) to account for
        # moving from disparite stratum to preceding common stratum
        last_common_idx = (
            first_disparite_idx - collision_implausibility_threshold
        )
        last_common_rank = first.GetRankAtColumnIndex(
            last_common_idx,
        )
        assert last_common_idx == last_common_rank
        return last_common_rank
    else:
        # no common strata between first and second
        # or not enough common strata to discount the possibility all
        # are spurious collisions with respect to the given confidence
        # level; conservatively conclude there is no common ancestor
        return None
